#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the runtime scope for the Macal virtual machine
#
# The primary function of this class is to provide a scope for the Macal virtual machine.
# It is used to store the values of the pre-defined variables, if there are any.

from __future__ import annotations
from typing import Tuple, Optional, Any, List
from .ast_nodetype import AstNodetype
from .conversion import typeFromValue
from .ct_variable import CTVariable

class RuntimeScope:
    def __init__(self) -> RuntimeScope:
        self.name: str = "runtime"
        self.variables: List[CTVariable] = []

    
    
    def DefineVariable(self, name: str, stack_offset: int) -> CTVariable:
        var = CTVariable(name, self)
        var.stack_offset = stack_offset
        self.variables.append(var)
        return var

    
    
    def GetVariable(self, name: str) -> Optional[CTVariable]:
        return next((var for var in self.variables if var.name == name), None)
        
    

    def SetVariable(self, name: str, value: Any) -> None:
        var = self.GetVariable(name)
        if var is None:
            raise Exception(f"Runtime Error: Variable {name} does not exist.")
        var.value = value
        var.node_type = typeFromValue(value)






#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# this is the bytecode library for the compiler and the scope.

from __future__ import annotations
from .conversion import convertToHexAddr

class BytecodeLibrary:
    def __init__(self, name: str, address: int) -> BytecodeLibrary:
        self.name: str = name
        self.address: int = address

    def __str__(self) -> str:
        return f"BytecodeLibrary(name={self.name}, address={convertToHexAddr(self.address)})"

    def __repr__(self) -> str:
        return self.__str__()


# Filename:       | meraki_api_library_v1.py
# Author:         | Marco Caspers
# Version:        | 9.1.1
# Description:
#
# Version 8.0.3, rewrite for Macal.
#
# Version 9.0.0, detached from MacalLibrary class for Macal version 3.5 which supports external libraries.
# Use this file in conjunction with meraki.mcl
#
# Rewrite to remove dependancy on meraki_api_interface.py
#
# Version 9.1.0, Macal 4.0.3 release
#
# Version 9.1.1, fixes for Macal 4.0.75
#
# Version 9.1.2, fixes for Macal 4.0.82
#
# Version 9.1.4
#
# Moved ERROR_TRAFFIC_AND_VISIBILITY and ERROR_NOT_FOUND_TRAFFIC constants from ../utilities.py into this file.
# This is to fix failing imports. The other functions that were imported from utilities where just shims for imports from os.path and inspect
# So i moved those imports directly here.
#
# This was needed because the new method of installing the Meraki Agent on Check Mk v2.0 would cause the import to fail.
#

# Version 9.1.5
#
# Added GetApplianceVpnStats Meraki API call.
#

# Version 10.0.0
#
# Upgraded for Macal version 5.0
#
# Using new library interface of Macal, aka no interface at all :)
#

from typing import Any
import meraki
from meraki.exceptions import APIError
from os.path import dirname, exists
from inspect import getframeinfo, currentframe
from datetime import datetime
import time
import json


Dapi = None
LastErrorMessage = ''

LAST_REQUEST = datetime.now()
EXEC_DELAY = 0.21
DEFAULT_REQUEST_COUNT = 1
API_CALL_THROTTLE_DELAY = 3
API_CALL_THROTTLE_PORT_STATUS_DELAY = 6

ERROR_TRAFFIC_AND_VISIBILITY   = "Traffic Analysis with Hostname Visibility must be enabled on this network to retrieve traffic data."
ERROR_NOT_FOUND_TRAFFIC        = "ERROR 0x10: Traffic data for networkID {} was not found."



def loadJson(fileName):
	"""Loads the json file from a given filename/path it does check for the existence of the file, returns None if no file exists."""
	if (exists(fileName)):
		with open(fileName) as json_file:
			return json.load(json_file)
	return None


def RequestsThrottle(p_requestcount=DEFAULT_REQUEST_COUNT):
	"""Execution Throttler
		This is used to delay the execution of network requests so that the limit 
		for the total number of requests is not exceded.
		Exceeding the limit of requests per second will result in fatal error, 
		bad request responses by the API."""
	# taking this one away makes meraki_test.mcl about 2.3 times faster. 
	# From almost 1 second down to just shy of 0.4 seconds.
	global LAST_REQUEST
	timeElapsed = (datetime.now() - LAST_REQUEST).total_seconds()
	if (timeElapsed < (EXEC_DELAY * p_requestcount)):
		#time.sleep(EXEC_DELAY * p_requestcount - timeElapsed)
		pass
	LAST_REQUEST = datetime.now()


def FnInitDashboardApi(api_key):
	"""Implementation of FnInitDashboardapi function"""
	global Dapi
	global LastErrorMessage
	try:
		Dapi = meraki.DashboardAPI(api_key, suppress_logging = True)
		return True
	except APIError as error:
		LastErrorMessage = f"Init API: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Init API unhandled Exception: {ex}'
	return False

def GetApiVersion() -> Any:
	if Dapi:
		return Dapi._session._version
	return False



def GetLastErrorMessage() -> str:
	"""Implementation of GetLastErrorMessage function, just returns the last error message to the scripting engine."""
	global LastErrorMessage
	return LastErrorMessage


def GetOrganizations() -> None:
	"""get_app_Organizations()
		input
			Dapi 		- Dashboard API

		The API returns a list of organizations.

        https://developer.cisco.com/meraki/api-v1/#!get-organizations
	"""
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizations()
	except APIError as error:
		if "errors" in error.response:
			emsg = ""
			for msg in error.response["errors"]:
				emsg = f"{emsg}\n{msg}"
			LastErrorMessage = f"Get Organizations: {emsg}"
		LastErrorMessage = f"Get Organizations: {error.response}"
	except Exception as ex:
		LastErrorMessage = f"Get Organizations: Unhandled exception: {ex}"
	return False



def GetOrganization(org_id) -> None:
	"""get_app_Organizations()
		input
			Dapi 		- Dashboard API
			org_id      - Organization ID

		The API returns the information of the organization.

        https://developer.cisco.com/meraki/api-v1/#!get-organization
	"""
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganization(org_id)
	except APIError as error:
		if "errors" in error.response:
			emsg = ""
			for msg in error.response["errors"]:
				emsg = f"{emsg}\n{msg}"
			LastErrorMessage = f"Get Organization: {emsg}"
		LastErrorMessage = f"Get Organization: {error.response}"
	except Exception as ex:
		LastErrorMessage = f"Get Organization: Unhandled exception: {ex}"
	return False



def GetInventory(org_id) -> None:
	"""get_org_Inventory(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of organizations inventory
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-inventory-devices"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationInventoryDevices(org_id, total_pages = -1)
	except APIError as error:
		LastErrorMessage = f"Get Inventory: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Inventory: Unhandled Exception: {ex}'
	return False

def GetDevices(org_id) -> None:
	"""get_org_Devices(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of organizations devices
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-devices"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationDevices(org_id, total_pages = -1)
	except APIError as error:
		LastErrorMessage = f"Get Devices: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Devices: Unhandled Exception: {ex}'
	return False


def GetDevice(serial) -> None:
	"""get_dev_Device(Dapi, serial)

		Input:
	   		Dapi      - Dashboard API
	   		serial    - Device serial number

	   	Output:
	   		Get a device from the list of devices.
	   		https://developer.cisco.com/meraki/api-v1/#!get-device"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.devices.getDevice(serial)
	except APIError as error:
		LastErrorMessage = f"Get Device: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Device: Unhandled Exception: {ex}'
	return False



def GetDevicesStatuses(org_id) -> None:
	"""get_org_DevicesStatuses(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of statuses of organizations devices
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-devices-statuses"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationDevicesStatuses(org_id, total_pages = -1)
	except APIError as error:
		LastErrorMessage = f"Get Devices Statusses: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Devices Statusses: Unhandled Exception: {ex}'
	return False



def GetDevicesLatency(org_id) -> None:
	"""get_org_DevicesUplinksLossAndLatency(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of Uplink, packet loss and latency data for organizations devices
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-devices-uplinks-loss-and-latency"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationDevicesUplinksLossAndLatency(org_id, total_pages = -1)
	except APIError as error:
		LastErrorMessage = f"Get Device Uplink Loss and Latency: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Device Uplink Loss and Latency: Unhandled Exception: {ex}'
	return False


def GetDevicesLatencyEx(org_id) -> None:
	"""get_org_DevicesUplinksLossAndLatency(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of Uplink, packet loss and latency data for organizations devices
	   		This is the enhanced version that calculates the averages this is to reduce the clunk in the macal script.
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-devices-uplinks-loss-and-latency"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	#try:
	result = Dapi.organizations.getOrganizationDevicesUplinksLossAndLatency(org_id, total_pages = -1)
	#{'networkId': 'N_645140646620848439', 'serial': 'Q2JN-94W8-MN69', 'uplink': 'wan2', 'ip': '8.8.8.8', 'timeSeries': [{'ts': '2022-09-05T13:10:37Z', 'lossPercent': 0.0, 'latencyMs': 3.1}, {'ts': '2022-09-05T13:11:37Z', 'lossPercent': 0.0, 'latencyMs': 3.0}, {'ts': '2022-09-05T13:12:36Z', 'lossPercent': 0.0, 'latencyMs': 2.9}, {'ts': '2022-09-05T13:13:37Z', 'lossPercent': 0.0, 'latencyMs': 3.1}, {'ts': '2022-09-05T13:14:37Z', 'lossPercent': 0.0, 'latencyMs': 3.2}]}
	ret = []
	for item in result:
		rec = {}
		rec['networkId'] = item['networkId']
		rec['serial'] = item['serial']
		rec['uplink'] = item['uplink']
		rec['ip'] = item['ip']
		rec['lossPercent'] = 0.0
		rec['latencyMs'] = 0.0
		rec['timeSeries'] = item['timeSeries']
		countLoss    = len(item['timeSeries'])
		countLatency = len(item['timeSeries'])
		for ts in item['timeSeries']:
			if ts['lossPercent'] is None:
				countLoss -= 1
			else:
				rec['lossPercent'] += ts['lossPercent']
			if ts['latencyMs'] is None:
				countLatency -= 1
			else:
				rec['latencyMs'] += ts['latencyMs']
		if countLoss > 0:
			rec['lossPercent'] = rec['lossPercent'] / countLoss
		if countLatency > 0:
			rec['latencyMs'] = rec['latencyMs'] / countLatency
		ret.append(rec)
	return ret



def GetDeviceManagementInterface(serial) -> None:
	"""
		get_devices_managementInterface(Dapi, serial)

		Input:
			Dapi   - Dashboard API
	   		serial - Serial number of the device.

		Output:
			Return the management interface settings for a device
			https://developer.cisco.com/meraki/api-v1/#!get-device-management-interface

	"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.devices.getDeviceManagementInterface(serial)
	except APIError as error:
		LastErrorMessage = f"Get Device Management Interface Settings: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Device Management Interface Settings: Unhandled Exception: {ex}'
	return False



def GetConfigTemplates(org_id) -> None:
	"""get_org_ConfigTemplates(Dapi, org_id)

		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of configuration templates of this organization
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-config-templates"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationConfigTemplates(org_id)
	except APIError as error:
		LastErrorMessage = f"Get Config Templates: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Config Templates: Unhandled Exception: {ex}'
	return False

def GetNetworks(org_id) -> None:
	"""get_org_Networks(Dapi, org_id)
		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID

	   	Output:
	   		list of networks in the organization
	   		https://developer.cisco.com/meraki/api-v1/#!get-organization-networks"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationNetworks(org_id)
	except APIError as error:
		LastErrorMessage = f"Get Networks: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Networks: Unhandled Exception: {ex}'
	return False

def GetNetwork(network_id) -> None:
	"""get_org_Network(Dapi, network_id)
		Input:
	   		Dapi   - Dashboard API
	   		network_id - Network ID

	   	Output:
	   		Get a single network
	   		https://developer.cisco.com/meraki/api-v1/#!get-network"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.networks.getNetwork(network_id)
	except APIError as error:
		LastErrorMessage = f"Get Network: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Network: Unhandled Exception: {ex}'
	return False

def GetApplianceUplinkStatuses(org_id) -> None:
	"""get_app_UplinkStatuses(Dapi, org_id)
		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID
		Output:
		List the uplink status of every Meraki MX and Z series appliances in the organization
        https://developer.cisco.com/meraki/api-v1/#!get-organization-appliance-uplink-statuses"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getOrganizationApplianceUplinkStatuses(org_id, total_pages = -1)
	except APIError as error:
		LastErrorMessage = f"Get Organization Appliance Uplink Statuses: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Organization Appliance Uplink Statuses: Unhandled Exception: {ex}'
	return False

def GetApplianceUplinkStatus(org_id, serials = None) -> None:
	"""get_app_UplinkStatuses(Dapi, org_id, serials)
		Input:
	   		Dapi   - Dashboard API
	   		org_id - Organization ID
		Output:
		List the uplink status of every Meraki MX and Z series appliances in the organization
        https://developer.cisco.com/meraki/api-v1/#!get-organization-appliance-uplink-statuses"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getOrganizationApplianceUplinkStatuses(org_id, serials = serials)
	except APIError as error:
		LastErrorMessage = f"Get Organization Appliance Uplink Statuses: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Organization Appliance Uplink Statuses: Unhandled Exception: {ex}'
	return False


def GetLicenseOverview(org_id) -> None:
	"""get_org_LicensesOverview(Dapi, org_id)

		Input:
	   		Dapi       - Dashboard API
	   		org_id     - Organization ID

		Output:
			LicensesOverview for the organization.
			https://developer.cisco.com/meraki/api-v1/#!get-organization-licenses-overview"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.organizations.getOrganizationLicensesOverview(org_id)
	except APIError as error:
		LastErrorMessage = f"Get License Overview: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get License Overview: Unhandled Exception: {ex}'
	return False



def GetAppPerformance(serial) -> None:
	"""get_app_Performance(Dapi, serial)
		Input:
	   		Dapi   - Dashboard API
	   		serial - Device Serial Number
		Output:
		Return the performance score for a single device. Only primary MX devices supported. If no data is available, a 204 error code is returned.
        https://developer.cisco.com/meraki/api-v1/#!get-device-appliance-performance"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getDeviceAppliancePerformance(serial)
	except APIError as error:
		if error.status == 400:
			LastErrorMessage = f'Get device appliance performance: {error.operation}: {error.status}, {serial} {error.message["errors"][0]}'
		elif error.status != 404:
			LastErrorMessage = f"Get device appliance performance: {error.operation}: {error.status}, {error.message}, {error.reason}"
		LastErrorMessage = f"Get device appliance performance Unknown error for {serial}: {error.status}, {error.response}"
	except Exception as ex:
		LastErrorMessage = f'Get device appliance performance Unhandled Exception: {ex}'
	return False



def GetNetworkTraffic(network_id, timespan=7200) -> None:
	"""get_app_NetworkTraffic(Dapi, network_id, T_timepsan=7200)

		Input:
	   		Dapi       - Dashboard API
	   		network_id - Network ID
	   		timespan   - Timespan in which to measure data, default 2 hours.

		Output:
			The traffic analysis data for this network.
            https://documentation.meraki.com/MR/Monitoring_and_Reporting/Hostname_Visibility
            ->Traffic Analysis with Hostname Visibility must be enabled on the network.
	        https://developer.cisco.com/meraki/api-v1/#!get-network-traffic"""
	global Dapi
	global LastErrorMessage
	print("@ This function is obsolete please use get_app_NetworkApplianceUplinkUsageHistory")
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	result = None
	try:
		return Dapi.networks.getNetworkTraffic(network_id, timespan = timespan)
	except APIError as error:
		if error.status == 400:
			if "errors" in error.message:
				if (len(error.message["errors"]) > 0):
					if (error.message["errors"][0] == ERROR_TRAFFIC_AND_VISIBILITY):
						LastErrorMessage = f"{ERROR_TRAFFIC_AND_VISIBILITY} (NetworkID: {network_id}"
			else:
				LastErrorMessage = f"{error.operation}: {error.status}, {error.message}, {error.reason}"
		elif error.status == 404:
			LastErrorMessage = ERROR_NOT_FOUND_TRAFFIC.format(network_id)
		else:
			LastErrorMessage = f"{error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Unhandled Exception: {ex}'
	return False



def GetNetworkEvents(network_id, product_type, start_after = -1) -> None:
	"""get_net_NetworkEvents()
		input:
			Dapi 		 - Dashboard API
			network_id   - Network ID
			product_Type - Product Type of the appliance
							  Valid types are wireless, appliance, switch, systemsManager, camera, and cellularGateway

			start_After  - The date/time from where to start collecting the data
							use -1 if you want to list all the events, regardless of the start date/time

		output:
			Gets a list of all network events that happened after the given date/time, for the specified product type, on the given network_id.
			https://developer.cisco.com/meraki/api-v1/#!get-network-events"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		if (start_after == -1):
			result = Dapi.networks.getNetworkEvents(network_id, productType = product_type, total_pages = 10)
		else:
			result = Dapi.networks.getNetworkEvents(network_id, productType = product_type, startingAfter = start_after)
		return result
	except APIError as error:
		LastErrorMessage = f"Get Network Events: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Network Events: Unhandled Exception: {ex}'
	return False


def GetOrgSecurityEvents(org_id, time_span = -1) -> None:
	"""get_app_OrganizationSecurityEvents()
		input
			Dapi 		- Dashboard API
			org_id 		- Organization ID
			time_span 	- Timespan over which to collect the data.
							Default is 5 minutes (300 seconds),
							Use -1 to list all data.

		List the security events for an organization**
        https://developer.cisco.com/meraki/api-v1/#!get-organization-appliance-security-events
	"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		if (time_span == -1):
			result = Dapi.appliance.getOrganizationApplianceSecurityEvents(org_id, total_pages = -1)
		else:
			result = Dapi.appliance.getOrganizationApplianceSecurityEvents(org_id, timespan = time_span)
		return result
	except APIError as error:
		LastErrorMessage = f"Get Security Events: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Security Events: Unhandled Exception: {ex}'
	return False

def GetSwitchPortStatuses(serial, time_span = -1) -> None:
	"""get_switch_DeviceSwitchPortsStatuses(Dapi, org_id)

		Input:
	   		Dapi     - Dashboard API
	   		serial   - Serial of the device
	   		timespan - timespan in seconds (300 = 5m default.)

	   	Output:
	   		list of statuses for ports for a specific switch.
	   		https://developer.cisco.com/meraki/api-v1/#!get-device-switch-ports-statuses"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_PORT_STATUS_DELAY)
	try:
		return Dapi.switch.getDeviceSwitchPortsStatuses(serial=serial, timespan = time_span)
	except APIError as error:
		LastErrorMessage = f"Get Switch Port Statuses: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Switch Port Statuses: Unhandled Exception: {ex}'
	return False



def GetSwitchPortStatusT0(serial, t0) -> None:
	"""get_switch_DeviceSwitchPortsStatuses(Dapi, org_id)

		Input:
	   		Dapi     - Dashboard API
	   		serial   - Serial of the device
	   		t0 - The beginning of the timespan for the data. The maximum lookback period is 31 days from today.

	   	Output:
	   		list of statuses for ports for a specific switch.
	   		https://developer.cisco.com/meraki/api-v1/#!get-device-switch-ports-statuses"""
	global Dapi
	global LastErrorMessage	
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.switch.getDeviceSwitchPortsStatuses(serial=serial, t0=t0)
	except APIError as error:
		LastErrorMessage = f"Get Switch Port Statuses T0: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Switch Port Statuses T0: Unhandled Exception: {ex}'
	return False


def GetApplianceUplinkUsageHistory(network_id) -> None:
	"""
		getNetworkApplianceUplinksUsageHistory.

		by default timespan   is 10 minutes.
		by default resolution is 1 minute.

		so by default a query returns 10 results per interface.

		The dimension is bytes.

		https://developer.cisco.com/meraki/api-v1/#!get-network-appliance-uplinks-usage-history
	"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	timespan = 600 # The timespan is 10 minutes, we can set the default here 
	               # because we don't allow changing of the default in this function call.
	try:
		upl = Dapi.appliance.getNetworkApplianceUplinksUsageHistory(network_id)
		nr = upl[0]['byInterface']
		for i, pt in enumerate(upl):
			if i > 0:
				inf = pt['byInterface']
				for f, intf in enumerate(inf):
					if nr[f]["sent"] is None:
						nr[f]["sent"] = 0
					if nr[f]["received"] is None:
						nr[f]["received"] = 0
					if intf is not None:
						if intf["sent"] is not None:
							nr[f]['sent'] += intf["sent"]
						if intf["received"] is not None:
							nr[f]['received'] += intf["received"]
		# Recalculate to bytes per second by deviding by the timespan
		for f in nr:
			f["total"] = (f["sent"] + f["received"]) / timespan
			f["sent"] = f["sent"] / timespan
			f["received"] = f["received"] / timespan
		return nr
	except APIError as error:
		LastErrorMessage = f"Get Appliance Uplink Usage History: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Appliance Uplink Usage History: Unhandled Exception: {ex}'
	return False



def GetApplianceTrafficShapingUplinkBandwidth(network_id) -> None:
	"""
    		**Returns the uplink bandwidth settings for your MX network.**
    		https://developer.cisco.com/meraki/api-v1/#!get-network-appliance-traffic-shaping-uplink-bandwidth
    		- networkId (string): (required)
    """
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getNetworkApplianceTrafficShapingUplinkBandwidth(network_id)
	except APIError as error:
		LastErrorMessage = f"Get Appliance Traffic Shaping Uplink Bandwidth: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Appliance Traffic Shaping Uplink Bandwidth: Unhandled Exception: {ex}'
	return False

def GetWirelessStatus(serial) -> None:
	"""https://developer.cisco.com/meraki/api/#!get-network-device-wireless-status"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.wireless.getDeviceWirelessStatus(serial)
	except APIError as error:
		LastErrorMessage = f"Get Wireless Status: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Wireless Status: Unhandled Exception: {ex}'
	return False


def GetApplianceVpnStats(org_id) -> None:
	"""https://developer.cisco.com/meraki/api-latest/#!get-organization-appliance-vpn-stats"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getOrganizationApplianceVpnStats(org_id, total_pages='all')
	except APIError as error:
		LastErrorMessage = f"Get VPN stats: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get VPN stats: Unhandled Exception: {ex}'
	return False

def GetApplianceVpnStatuses(org_id) -> None:
	"""https://developer.cisco.com/meraki/api-latest/#!get-organization-appliance-vpn-statuses"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.appliance.getOrganizationApplianceVpnStatuses(org_id, total_pages='all')
	except APIError as error:
		LastErrorMessage = f"Get VPN stats: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get VPN stats: Unhandled Exception: {ex}'
	return False

def Output(*params) -> None:
	"""Implementation of Output function. output => (params args)"""
	if len(params) != 2:
		raise Exception(f"Output requires 2 arguments got ({len(params)}).")
	v = params[1]
	if (v is not None and v != '' and v != {} and v != [] and v != "nil"):
		print(params[0])
		print(v)



def OutputOnCondition(*params) -> None:
	"""Implementation of OutputOnCondition function"""
	if params[0] is False: return
	if len(params) != 3:
		raise Exception(f"OutputOnCondition requires 3 arguments got ({len(params)}).")
	params = params[1:]
	v = params[1]
	if (v is not None and v != '' and v != {} and v != [] and v != "nil"):
		print(params[0])
		print(v)

def Header(*params) -> None:
	"""Implementation of Header function"""
	print("<<<meraki_devices>>>")
	v = params[1]
	if (params[0] == True):
		print(f"[{v}_demo]")
	else:
		print(f"[{v}]")
	model = params[2]
	serial = params[3]
	template = params[4]
	if model is None or model ==  "nil": model = ""
	if serial is None or serial == "nil": serial = ""
	if template is None or template == "nil": template = ""
	print(f"{model} - {serial} - {template}")


def ProductInfo(*params) -> None:
	"""Implementation of Header function"""
	print("<<<meraki_product_info>>>")
	model = params[0]
	serial = params[1]
	template = params[2]
	if model is None or model == "nil": model = ""
	if serial is None or serial == "nil": serial = ""
	if template is None or template == "nil": template = ""
	print(f"{model} - {serial} - {template}")

def DeviceHeader(*params) -> None:
	"""Implementation of DeviceHeader function"""
	if params[0] == True:
		print(f"<<<<{params[1]}_demo>>>>")
	else:
		print(f"<<<<{params[1]}>>>>")

def ReadFilters(customer) -> None:
	"""Implementation of DeviceHeader function"""
	if not customer:
		print("@customer did not exist")
		return None
	# Valid types are wireless, appliance, switch, systemsManager, camera, and cellularGateway
	filters = {}
	filters_base = customer["network_event_filters"][0]
	filename  = getframeinfo(currentframe()).filename   # get the filename/path where this script is located.
	filepath  = dirname(filename)			        # get the path of this file.
	# if we get the externals subfolder back, cut it off!
	if filepath.endswith("/external"):
		filepath = filepath[0:-9]
	if "switch" in filters_base:
		filename =  f"""{filepath}{filters_base["switch"]}"""
		filters["switch"] = loadJson(filename)
	if "wireless" in filters_base:
		filename =  f"""{filepath}{filters_base["wireless"]}"""
		filters["wireless"] = loadJson(filename)
	if "appliance" in filters_base:
		filename =  f"""{filepath}{filters_base["appliance"]}"""
		filters["appliance"] = loadJson(filename)
	if "systemsManager" in filters_base:
		filename =  f"""{filepath}{filters_base["systemsManager"]}"""
		filters["systemsManager"] = loadJson(filename)
	if "camera" in filters_base:
		filename =  f"""{filepath}{filters_base["camera"]}"""
		filters["camera"] = loadJson(filename)
	if "cellularGateway" in filters_base:
		filename =  f"""{filepath}{filters_base["cellularGateway"]}"""
		filters["cellularGateway"] = loadJson(filename)
	return filters

def FlattenRecord(rec) -> None:
	"""Implementation of FlattenRecord function"""
	result = ""
	for (key, value) in rec.items():
		result = "{},'{}','{}'".format(result, key, value)
	return result[1:]


def FlattenTuple(tup) -> None:
	"""Implementation of FlattenTuple function"""
	result = []
	for item in tup:
		if isinstance(item, str):
			result.append(f"'{item}'")
		else:
			result.append(item)
	return result


def GetSensorReadingsHistory(org_id) -> None:
	"""https://developer.cisco.com/meraki/api-v1/#!get-organization-sensor-readings-history"""
	global Dapi
	global LastErrorMessage
	RequestsThrottle(API_CALL_THROTTLE_DELAY)
	try:
		return Dapi.sensor.getOrganizationSensorReadingsHistory(org_id)
	except APIError as error:
		LastErrorMessage = f"Get Sensor Readings History: {error.operation}: {error.status}, {error.message}, {error.reason}"
	except Exception as ex:
		LastErrorMessage = f'Get Sensor Readings History: Unhandled Exception: {ex}'
	return False


def StrToInt(str) -> None:
	"""https://developer.cisco.com/meraki/api-v1/#!get-organization-sensor-readings-history"""
	global LastErrorMessage
	try:
		return int(str)
	except Exception as ex:
		LastErrorMessage = f"StrToInt(): {ex}"




# Product:   Macal 2
# Author:    Marco Caspers
# Date:      16-10-2023
#

"""Macal system library implementation"""

from typing import Any
import platform
import os
import sys

def Console(*args, **kwargs):
    print(*args, **kwargs)


def Array(*args) -> None:
    arr = []
    for arg in args:
        arr.append(arg)
    return arr
    

def RecordHasField(record: dict, fieldname: str) -> bool:
    return fieldname in record

def GetPlatform() -> None:
    return platform.system()

def RecordItems(var : dict) -> list:
    """Implementation of Items function used in conjunction with foreach for iterating over records.  Items returns key/value pairs."""
    return [{key: value} for key, value in var.items()]

def RecordItemKey(var : dict) -> str:
    """Implementation of Key function used in conjunction the Items function that returns key/value pairs. Key returns the key part of a key value pair."""
    for k, _ in var.items(): #there are different ways, but this is by far the most simple and safe way to do it.
        return k

def RecordKeys(var: dict) -> list:
    """Implementation of Keys function used in conjunction the Items function that returns key/value pairs. Key returns the key part of a key value pair."""
    return [k for k in var.keys()]

def RecordItemValue(var: dict) -> Any:
    """Implementation of Value function used in conjunction the Items function that returns key/value pairs. Value returns the value part of a key value pair."""
    for _, v in var.items(): #there are different ways, but this is by far the most simple and safe way to do it.
        return v

def RecordValues(var: dict) -> list:
    """Implementation of Value function used in conjunction the Items function that returns key/value pairs. Value returns the value part of a key value pair."""
    return [v for v in var.values()]

def GetEnv(varname: str) -> str:
    """Returns the value of an environment variable"""
    return os.getenv(varname)

def SetEnv(varname: str, value: str) -> None:
    """Sets the value of an environment variable"""
    os.environ[varname] = value

def GetArgs() -> list:
    """Returns a list of arguments passed to the program"""
    return sys.argv

def GetArg(index: int) -> str:
    """Returns a single argument passed to the program"""
    return sys.argv[index]

def GetArgCount() -> int:
    """Returns the number of arguments passed to the program"""
    return len(sys.argv)

def GetType(var: Any) -> str:
    """Returns the type of a variable as a string"""
    tts = str(type(var)).split("'")[1]
    if tts == 'NoneType':
        return 'nil'
    if tts == 'bool':
        return 'boolean'
    if tts == 'int':
        return 'integer'
    if tts == 'list':
        return 'array'
    if tts == 'dict':
        return 'record'
    if tts == 'str':
        return 'string'
    return tts



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytecode register.

from __future__ import annotations
import sys
from typing import Any

from .ast_nodetype import AstNodetype
from .conversion import compareTypes, typeFromValue

class BytecodeRegister:
    def __init__(self, name: str, opcode: int) -> BytecodeRegister:
        self.name: str = name                  # register name
        self.opcode: int = opcode              # register opcode, this is used in the vm to identify the register
        self.value: Any = 0
        self.value_type: AstNodetype = AstNodetype.INTEGER
        self._do_raise: bool = True             # raise an exception if an error occurs

    def reset(self) -> None:
        self.value_type = AstNodetype.INTEGER
        self.value = 0

    def error(self, message: str) -> None:
        msg = f"Register Error: {message}"
        if self._do_raise:
            raise Exception(msg)
        print(msg)
        sys.exit(1)


    def __str__(self) -> str:
        return f"{self.name} {self.value_type.name} {self.value} {self.opcode}"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: BytecodeRegister) -> BytecodeRegister:
        self.value = self.value_type == other.value_type and self.value == other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __ne__(self, other: BytecodeRegister) -> BytecodeRegister:
        self.value = self.value_type != other.value_type or self.value != other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __lt__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot compare {self.value_type} with {other.value_type}")
        self.value = self.value < other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __le__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot compare {self.value_type} with {other.value_type}")
        self.value = self.value <= other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __gt__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot compare {self.value_type} with {other.value_type}")
        self.value = self.value > other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __ge__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot compare {self.value_type} with {other.value_type}")
        self.value = self.value >= other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __expand_to_float__(self, other: BytecodeRegister) -> bool:
        # expand to float if one of the types is float.
        if ((self.value_type == AstNodetype.FLOAT or other.value_type == AstNodetype.FLOAT) and
           (self.value_type == AstNodetype.INTEGER or other.value_type == AstNodetype.INTEGER)):
            self.value_type = AstNodetype.FLOAT
            self.value = float(self.value)
            other.value = float(other.value)
            return True
        return False

    def toStr(self, value: Any) -> str:
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if value is None or value == AstNodetype.NIL:
            return 'nil'
        return str(value)

    def __expand_to_string__(self, other: BytecodeRegister) -> bool:
        if (self.value_type in (AstNodetype.STRING, AstNodetype.NIL,
                          AstNodetype.STRING_INTERPOLATION_STRING_PART,
                          AstNodetype.STRING_INTERPOLATION_END)
                          or other.value_type in
                          (AstNodetype.STRING, AstNodetype.NIL,
                           AstNodetype.STRING_INTERPOLATION_STRING_PART,
                           AstNodetype.STRING_INTERPOLATION_END)):
            self.value_type = AstNodetype.STRING
            self.value = self.toStr(self.value)
            other.value = self.toStr(other.value)
            return True
        return False

    def __add__(self, other: BytecodeRegister) -> BytecodeRegister:
        if not compareTypes(self.value_type, other.value_type):
            if not (self.__expand_to_float__(other) or self.__expand_to_string__(other)):
                self.error(f"Cannot add {self.value_type} with {other.value_type}")
        self.value += other.value
        return self

    def __sub__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            if not self.__expand_to_float__(other):
                self.error(f"Cannot subtract {self.value_type} with {other.value_type}")
        self.value -= other.value
        return self

    def __mul__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            if not self.__expand_to_float__(other):
                self.error(f"Cannot multiply {self.value_type} with {other.value_type}")
        self.value *= other.value
        return self

    def __truediv__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            if not self.__expand_to_float__(other):
                self.error(f"Cannot divide {self.value_type} with {other.value_type}")
        self.value /= other.value
        self.value_type = typeFromValue(self.value)
        return self

    def __mod__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            if not self.__expand_to_float__(other):
                self.error(f"Cannot modulus {self.value_type} with {other.value_type}")
        self.value %= other.value
        return self

    def __pow__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            if not self.__expand_to_float__(other):
                self.error(f"Cannot power {self.value_type} with {other.value_type}")
        self.value **= other.value
        return self

    def __and__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot AND {self.value_type} with {other.value_type}")
        self.value = self.value and other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __or__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot OR {self.value_type} with {other.value_type}")
        self.value = self.value or other.value
        self.value_type = AstNodetype.BOOLEAN
        return self

    def __xor__(self, other: BytecodeRegister) -> BytecodeRegister:
        if self.value_type != other.value_type:
            self.error(f"Cannot XOR {self.value_type} with {other.value_type}")
        if self.name == other.name:
            self.value = int(0)
            self.value_type = AstNodetype.INTEGER
            return self
        self.value = self.value ^ other.value
        return self

    def __bool__(self) -> BytecodeRegister:
        if self.value_type != AstNodetype.BOOLEAN:
            self.error(f"Cannot invert {self.value_type}")
        self.value = not self.value
        return self

    def __neg__(self) -> BytecodeRegister:
        if self.value_type != AstNodetype.INTEGER and self.value_type != AstNodetype.FLOAT:
            self.error(f"Cannot negate {self.value_type}")
        self.value = -self.value
        return self

    def fromRegister(self, register: BytecodeRegister) -> None:
        self.value_type = register.value_type
        self.value = register.value

    def set(self, type: AstNodetype, value: Any) -> None:
        self.value_type = type
        self.value = value




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# Node type definitions for Macal

from enum import Enum, auto

class AstNodetype(Enum):
    UNDEFINED = 127

    # types used at runtime
    INTEGER = auto()      # 128
    FLOAT = auto()        # 129
    STRING = auto()       # 130
    BOOLEAN = auto()      # 131
    ARRAY = auto()        # 132
    RECORD = auto()       # 133
    VARIABLE   = auto()   # 134
    FUNCTION   = auto()   # 135
    TYPE = auto()         # 136
    NIL = auto()          # 137
    LIBRARY = auto()      # 138
    LABEL = auto()        # 139
    BC_METADATA = auto()  # 130
    STRING_INTERPOLATION_STRING_PART = auto()# 131
    STRING_INTERPOLATION_END = auto()        # 132

    VM_INSTRUCTION = auto()
    # Token types
    IDENTIFIER = auto()
    WHITESPACE = auto()
    COMMENT = auto()
    EOF = auto()
    # variable types
    BOOLEAN_TRUE = auto()
    BOOLEAN_FALSE = auto()
    STRING_INTERPOLATION_START = auto()
    NEW_RECORD = auto()
    NEW_ARRAY = auto()
    INDEX = auto()
    ANY = auto()
    PARAMS = auto()
    # Operator type LexTokens
    OPERATOR_ARROW = auto()
    OPERATOR_ADDITION = auto()
    OPERATOR_SUBTRACTION = auto()
    OPERATOR_MULTIPLICATION = auto()
    OPERATOR_DIVISION = auto()
    OPERATOR_POWER = auto()
    OPERATOR_MODULUS = auto()
    OPERATOR_ASSIGNMENT = auto()
    OPERATOR_INCREMENT = auto()
    OPERATOR_DECREMENT = auto()
    OPERATOR_ASSIGNMENT_INC = auto()
    OPERATOR_ASSIGNMENT_DEC = auto()
    OPERATOR_ASSIGNMENT_MUL = auto()
    OPERATOR_ASSIGNMENT_DIV = auto()
    OPERATOR_ASSIGNMENT_MOD = auto()
    OPERATOR_ASSIGNMENT_POW = auto()
    OPERATOR_ASSIGNMENT_NEG = auto()
    # Comparison type LexTokens
    COMPARETOR_EQUAL = auto()
    COMPARETOR_NOT_EQUAL = auto()
    COMPARETOR_LESS_THAN = auto()
    COMPARETOR_LESS_THAN_EQUAL = auto()
    COMPARETOR_GREATER_THAN = auto()
    COMPARETOR_GREATER_THAN_EQUAL = auto()
    AND_STATEMENT = auto()
    OR_STATEMENT = auto()
    NOT_STATEMENT = auto()
    XOR_STATEMENT = auto()
    BITWISE_AND = auto()
    BITWISE_OR = auto()
    BITWISE_NOT = auto()
    # Punctuation LexTokens
    LEFT_PAREN = auto()
    RIGHT_PAREN = auto()
    LEFT_BRACKET = auto()
    RIGHT_BRACKET = auto()
    LEFT_BRACE = auto()
    RIGHT_BRACE = auto()
    COMMA = auto()
    SEMICOLON = auto()
    COLON = auto() # not implemented
    DOT = auto() # not implemented
    # AST Node types
    PROGRAM    = auto()
    BLOCK     = auto()
    END_BLOCK = auto()
    FUNCTION_CALL = auto()
    FUNCTION_DEFINITION = auto()

    IF_STATEMENT = auto()
    ELIF_STATEMENT_LIST = auto()
    ELIF_STATEMENT = auto()
    ELSE_STATEMENT = auto()
    FOREACH_STATEMENT = auto()
    WHILE_STATEMENT = auto()
    CONTINUE_STATEMENT = auto()
    BREAK_STATEMENT = auto()
    RETURN_STATEMENT = auto()
    SELECT_STATEMENT = auto()
    SELECT_FIELD = auto()
    DISTINCT_STATEMENT = auto()
    AS_STATEMENT = auto()
    FROM_STATEMENT = auto()
    WHERE_STATEMENT = auto()
    ORDER_STATEMENT = auto()
    MERGE_STATEMENT = auto()
    INTO_STATEMENT = auto()
    HALT_STATEMENT = auto()
    INCLUDE_STATEMENT = auto()
    EXTERNAL_STATEMENT = auto()
    PRINT_STATEMENT = auto()
    CONST_STATEMENT = auto()
    SWITCH_STATEMENT = auto()
    CASE_STATEMENT = auto()
    DEFAULT_STATEMENT = auto()
    ASSIGNMENT_STATEMENT = auto()
    TYPE_STATEMENT = auto()
    IS_STRING_STATEMENT = auto()
    IS_INT_STATEMENT = auto()
    IS_FLOAT_STATEMENT = auto()
    IS_BOOL_STATEMENT = auto()
    IS_RECORD_STATEMENT = auto()
    IS_ARRAY_STATEMENT = auto()
    IS_FUNCTION_STATEMENT = auto()
    IS_NIL_STATEMENT = auto()
    IS_TYPE_STATEMENT = auto()

    # expressions
    EXPRESSION = auto()
    INDEXED_VARIABLE_EXPRESSION = auto()
    VARIABLE_EXPRESSION = auto()
    FUNCTION_CALL_EXPRESSION = auto()
    VARIABLE_FUNCTION_CALL_EXPRESSION = auto()
    LIBRARY_EXPRESSION = auto()
    LITERAL_EXPRESSION = auto()
    LITERAL_ARRAY_EXPRESSION = auto()
    LITERAL_RECORD_EXPRESSION = auto()
    UNARY_EXPRESSION = auto()
    BINARY_EXPRESSION = auto()
    TERNARY_EXPRESSION = auto() # not implemented

    # misc
    JUMP_TABLE = auto()
    PRINT = auto()

    # misc 2
    BC_VERSION = auto() # 250!

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

if max([e.value for e in AstNodetype]) > 255:
    raise Exception("Too many AstNodetypes.  Max is 255.")


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Interactive Line Interpreter class

from .macal_lexer import Lexer
from .macal_parser import Parser
from .macal_compiler import MacalCompiler
from .macal_decompiler import MacalDecompiler
from .__about__ import __version__, __author__, __author_email__, __license__, __copyright__

class MacalInteractive:
    def __init__(self) -> None:
        self.lexer: Lexer = Lexer()
        self.parser: Parser = Parser()
        self.compiler: MacalCompiler = MacalCompiler()
        self.decompiler: MacalDecompiler = MacalDecompiler(None, True)
        self.do_except: bool = True

    def ReadLine(self) -> str:
        return input(">>> ")

    def ExecuteLine(self, text: str) -> None:
        tokens = self.lexer.lex(text)
        print(tokens)
        ast = self.parser.parse_interactive(tokens)
        print(ast)
        start = self.compiler.cs.rip.value
        self.compiler.cs.flags.trap = True
        self.compiler.compile_interactive(ast)
        end = self.compiler.cs.rip.value
        print("start, end = ", start, end)
        self.decompiler.vm.memory = self.compiler.cs.memory[start:end]
        self.compiler.cs.rip.value = start
        self.decompiler.Decompile()
        while self.compiler.cs.rip.value < end:
            self.compiler.cs.Execute()
            self.compiler.cs.Halted = False
        self.compiler.cs.flags.trap = False

    def RunStep(self) -> bool:
        text = self.ReadLine()
        if text is None or text.strip() == "":
            return True
        if text == "exit":
            return False
        if text == "credits":
            print("Product:   Macal")
            print("Author:    ", __author__)
            print("Email:     ", __author_email__)
        elif text == "license":
            print("License:   ", __license__)
        elif text == "copyright":
            print("Copyright:   ", __copyright__)
        elif text == 'reset':
            self.compiler.cs.Reset()
        else:
            self.ExecuteLine(text.strip())
        return True

    def Run(self) -> None:
        print("Macal Interactive Interpreter v" + __version__)
        print("Type 'exit' to exit. Type 'copyright', 'credits' or 'license' for more information.")
        self.compiler.cs.Reset()
        while True:
            if self.do_except:
               if not self.RunStep():
                  break
            else:
                try:
                  if not self.RunStep():
                      break
                except Exception as e:
                    print(e)
                    continue



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the function call statement node for the AST.

from __future__ import annotations
from typing import List, Union, Any, Optional
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_indexed_variable_expression import AstNodeIndexedVariableExpression

class AstNodeFunctionCallStatement(AstNodeStatement):
    def __init__(self, ident: Union[LexToken, AstNodeIndexedVariableExpression], args: List[AstNodeExpression]) -> AstNodeFunctionCallStatement:
        self.args: List[AstNodeExpression] = args
        self.index: Optional[List[Any]] = None
        self.var_expr: Optional[AstNodeIndexedVariableExpression] = None
        if isinstance(ident, AstNodeIndexedVariableExpression):
            super().__init__(ident._token)
            self.index = ident.index
            self.var_expr = ident
        else:
            super().__init__(ident)
        self.expr_type = AstNodetype.FUNCTION_CALL


    @property
    def name(self) -> str:
        return self.token.value

    def __str__(self) -> str:
        return f"AstNodeFunctionCallExpression({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}{self}")
        if self.args:
            for arg in self.args:
                arg.tree(indent + '    ')
        else:
            print()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# This is a wrapper of a python module for the library system.

from __future__ import annotations

from typing import Any

class ModuleInfo:
    def __init__(self, name: str, module: Any) -> ModuleInfo:
        self.name = name
        self.module = module
        import inspect
        members = inspect.getmembers(module)
        object_dict = {name: obj for name, obj in members}
        self.functions = {name: obj for name, obj in object_dict.items() if inspect.isfunction(obj)}

    def __str__(self) -> str:
        return f"{self.name}"

    def __repr__(self) -> str:
        return self.__str__()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for a variable

from __future__ import annotations
from typing import Optional
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeVariable(AstNodeExpression):
    def __init__(self, token: LexToken) -> AstNodeVariable:
        super().__init__(token)
        self.expr_type = AstNodetype.VARIABLE
        self.name: str = token.value
        self.value: Optional[AstNodeExpression] = None

    def __str__(self) -> str:
        return f"AstNodeVariable({self.name}, {self.value})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}Variable:")
        print(f"{indent}    {self.name}")
        if self.value is not None:
            self.value.tree(indent + '    ')


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# This is a method to bypass the limitation of Python that you can't pass a variable by reference.
# In this case we need to pass an address by reference, because the action that we take in the function
# needs to increase the address by the size of the data type that we are reading, so that the next time
# we read the address will be the next address in memory.

from __future__ import annotations
from .conversion import convertToHexAddr

class Address:
    def __init__(self, address: int = 0) -> Address:
        self.address: int = address

    def __str__(self) -> str:
        return f"{convertToHexAddr(self.address)}"

    def __repr__(self) -> str:
        return f"{convertToHexAddr(self.address)}"

    def __add__(self, other: int) -> Address:
        self.address += other
        return self

    def __sub__(self, other: int) -> Address:
        self.address -= other
        return self


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the bytcode compiler for Macal

from __future__ import annotations
import sys
import os
from typing import List, Optional
from .ast_nodetype import AstNodetype
from .ast_node_statement import AstNodeStatement
from .ast_node_expression import AstNodeExpression
from .ast_node_binary_expression import AstNodeBinaryExpression
from .ast_node_unary_expression import AstNodeUnaryExpression
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_node_program import AstNodeProgram
from .ast_node_block import AstNodeBlock
from .ast_node_if_statement import AstNodeIfStatement
from .ast_node_elif_statement import AstNodeElifStatement
from .ast_node_else_statement import AstNodeElseStatement
from .ast_node_assignment import AstNodeAssignmentStatement
from .ast_node_indexed_variable_expression import AstNodeIndexedVariableExpression
from .ast_node_function_call_expression import AstNodeFunctionCallExpression
from .ast_node_variable_expression import AstNodeVariableExpression
from .ast_node_variable_function_call_expression import AstNodeVariableFunctionCallExpression
from .ast_node_print_statement import AstNodePrintStatement
from .ast_node_break_statement import AstNodeBreakStatement
from .ast_node_continue_statement import AstNodeContinueStatement
from .ast_node_return_statement import AstNodeReturnStatement
from .ast_node_halt_statement import AstNodeHaltStatement
from .ast_node_library_expression import AstNodeLibraryExpression
from .ast_node_include_statement import AstNodeIncludeStatement
from .ast_node_while_statement import AstNodeWhileStatement
from .ast_node_foreach_statement import AstNodeForeachStatement
from .ast_node_function_definition import AstNodeFunctionDefinition
from .ast_node_function_call_statement import AstNodeFunctionCallStatement
from .ast_node_function_parameter import AstNodeFunctionParameter
from .ast_node_type_statement import AstNodeTypeStatement
from .ast_node_istype import AstNodeIsType
import pyximport; pyximport.install()
from .cmacal_vm import MacalVm
from .macal_instructions import Opcode
from .bytecode_optimizer import BytecodeOptimizer
from .compiler_scope import CompilerScope
from .bytecode_register import BytecodeRegister
from .ct_function import CTFunction
from .ct_variable import CTVariable
from .lex_token import LexToken
from .macal_select_compiler import MacalSelectCompiler
from .macal_lexer import Lexer
from .macal_parser import Parser
from .switch_case_table import SwitchCaseTable
from .config import SearchPath
from .macal_instruction_emitter import MacalInstructionEmitter

# Function stack layout:

# -6 PARAMETER 0
# -5 PARAMETER 1
# -4 PARAMCOUNT
# -3 RETURN VALUE
# -2 RETURN ADDRESS
# -1 RBP

FUNCTION_CALL_PARAMETER_STACK_OFFSET = -5
FUNCTION_CALL_PARAMETER_COUNT_STACK_OFFSET = -4
FUNCTION_CALL_RETURN_VALUE_STACK_OFFSET = -3
FUNCTION_CALL_RETURN_ADDRESS_STACK_OFFSET = -2
FUNCTION_CALL_PREVIOUS_RBP_STACK_OFFSET = -1

class MacalCompiler:
    def __init__(self, program: Optional[AstNodeProgram] = None, output_path: Optional[str] = None, verbose: bool = False, debug: bool = False, reserved_vars: Optional[List[str]] = None) -> MacalCompiler:
        self.program: Optional[AstNodeProgram] = program
        self.output_path: Optional[str] = output_path
        self.output_file: Optional[str] = None
        self.cs: MacalVm = MacalVm() # Code segment
        self.emitter: MacalInstructionEmitter = MacalInstructionEmitter(self.cs.memory, self.cs.rip, self.cs.rsp, self.cs.rbp)
        self.scope: CompilerScope = CompilerScope("global", None)
        self.do_raise: bool = True
        self.debug = debug
        compiler_path = os.path.dirname(__file__)
        lib_path = os.path.join(compiler_path, "lib")
        if lib_path not in SearchPath:
            SearchPath.append(lib_path)
        if compiler_path not in SearchPath:
            SearchPath.append(compiler_path)
        self.reserved_vars = reserved_vars
        self.verbose: bool = verbose



    def error(self, message: str) -> None:
        msg = f"Compiler Error: {message}"
        if self.do_raise:
            raise Exception(msg)
        print(msg)
        sys.exit(1)



    def compile(self) -> None:
        self.cs.Reset()
        # we need to start at an offset of 2 because exit code
        # and return address zero are on the stack and not to be touched.
        self.cs.rsp.value = 2
        self.cs.rbp.value = 2
        self.compile_program(self.program, self.scope)
        self.__resolve_jumps_and_labels__(self.scope)
        self.cs.memory = self.emitter.memory # fix the memory problem...
        #opt = BytecodeOptimizer(self.cs, self.scope)
        #opt.process()
        if self.verbose:
            self.compile_datasegment(self.scope)
        if len(self.scope.variables) > 0:
            self.compile_reserved_segment(self.scope)
        self.cs.Save(self.output_path, self.cs.rip.value)



    def compile_reserved_segment(self, scope: CompilerScope) -> None:
        pvars = []
        for var in scope.variables:
            pvars.append((var.name, var.stack_offset))
        metadata = {
            "VARIABLES": [],
            "FUNCTIONS": [],
            "LIBRARIES": [],
            "LABELS": [],
            "JUMPS": [],
            "RESERVED": pvars
        }
        self.emitter.RESERVEDS(metadata)



    def compile_interactive(self, ast: AstNodeProgram) -> None:
        self.compile_block(ast, self.scope)
        self.__resolve_jumps_and_labels__(self.scope)
        if self.verbose:
            self.compile_datasegment(self.scope)



    def compile_reserved_variables(self, scope: CompilerScope) -> None:
        """ This adds pre-defined variables onto the stack."""
        if self.reserved_vars is None:
            return
        offset = self.cs.rsp.value
        for var in self.reserved_vars:
            scope.define_variable(var.strip(), offset)
            offset += 1
        self.emitter.RESERVE(len(self.reserved_vars))
        self.cs.rsp.value += len(self.reserved_vars)



    def __resolve_jumps_and_labels__(self, scope: CompilerScope) -> None:
        for jump in scope.jump_table:
            jaddr: int = jump.address
            laddr: int = scope.labels[jump.label].address
            op = self.emitter.memory[jaddr]
            new_op = (op[0], laddr)
            self.emitter.memory[jaddr] = new_op

        for switch_case_table in scope.switch_jump_tables:
            for case, label in switch_case_table.labels.items():
                laddr: int = scope.labels[label].address
                switch_case_table.labels[case] = laddr

            op = self.emitter.memory[switch_case_table.address]
            new_op = (op[0], op[1], op[2], switch_case_table.labels)
            self.emitter.memory[switch_case_table.address] = new_op



    def compile_program(self, program: AstNodeProgram, scope: CompilerScope) -> None:
        label_main = self.scope.get_new_label("main")
        label_exit = self.scope.get_new_label("exit")
        self.scope.set_label_address(label_main, self.cs.rip.value)
        main_scope = self.emit_scope(scope)
        self.compile_reserved_variables(scope)
        #print("Compiling program...")
        #print("RSP: ", self.cs.rsp.value)
        self.compile_block(program, scope)
        #print("Program end...")
        #print("RSP: ", self.cs.rsp.value)
        self.scope.set_label_address(label_exit, self.cs.rip.value)
        self.emit_scope_end(main_scope)
        self.emitter.RET()



    def compile_block(self, block: AstNodeBlock, scope: CompilerScope) -> None:
        for statement in block.statements:
            self.compile_statement(statement, scope)



    def compile_statement(self, statement: AstNodeStatement, scope: CompilerScope) -> None:
        if statement.expr_type == AstNodetype.IF_STATEMENT:
            self.compile_if_statement(statement, scope)
        elif statement.expr_type == AstNodetype.FOREACH_STATEMENT:
            self.compile_foreach_statement(statement, scope)
        elif statement.expr_type == AstNodetype.WHILE_STATEMENT:
            self.compile_while_statement(statement, scope)
        elif statement.expr_type == AstNodetype.CONTINUE_STATEMENT:
            self.compile_continue_statement(statement, scope)
        elif statement.expr_type == AstNodetype.BREAK_STATEMENT:
            self.compile_break_statement(statement, scope)
        elif statement.expr_type == AstNodetype.RETURN_STATEMENT:
            self.compile_return_statement(statement, scope)
        elif statement.expr_type == AstNodetype.SELECT_STATEMENT:
            self.compile_select_statement(statement, scope)
        elif statement.expr_type == AstNodetype.HALT_STATEMENT:
            self.compile_halt_statement(statement, scope)
        elif statement.expr_type == AstNodetype.INCLUDE_STATEMENT:
            self.compile_include_statement(statement, scope)
        elif statement.expr_type == AstNodetype.PRINT_STATEMENT:
            self.compile_print_statement(statement, scope)
        elif statement.expr_type == AstNodetype.SWITCH_STATEMENT:
            self.compile_switch_statement(statement, scope)
        elif statement.expr_type == AstNodetype.ASSIGNMENT_STATEMENT:
            self.compile_assignment_statement(statement, scope)
        elif statement.expr_type == AstNodetype.FUNCTION_CALL:
            self.compile_function_call(statement, scope)
        elif statement.expr_type == AstNodetype.FUNCTION_DEFINITION:
            self.compile_function_definition(statement, scope)
        elif statement.expr_type == AstNodetype.FUNCTION_CALL_EXPRESSION:
            self.compile_function_call_expression(statement, scope)
        else:
            self.error(f"Unknown statement type {statement.expr_type} at {statement.line}:{statement.column} scope({scope.name})")

# code writing helpers.

    def new_child_scope(self, name: str, scope: CompilerScope,
                        lbl_continue: Optional[int] = None,
                        lbl_end: Optional[int] = None ) -> CompilerScope:
        child: CompilerScope = scope.new_child(name)
        child.continue_label = lbl_continue
        child.break_label = lbl_end
        child.rbp_offset = scope.rbp_offset



    def emit_condition(self, label: int, conditional_jump: int, scope: CompilerScope) -> None:
        self.emitter.POP(self.cs.rax)                                # pop rax
        self.emitter.CMPRI(self.cs.rax, AstNodetype.BOOLEAN, True)   # cmpir rax, True
        self.emit_jump(conditional_jump, label, scope)   # jmpnz label



    def emit_new_label(self, name: str, scope: CompilerScope) -> int:
        lbl = scope.root.get_new_label(name)
        scope.root.set_label_address(lbl, self.cs.rip.value)
        return lbl



    def emit_scope(self, scope: CompilerScope) -> CompilerScope:
        self.emitter.PUSH(self.cs.rbp)
        self.emitter.MOVRR(self.cs.rbp, self.cs.rsp)
        child: CompilerScope = scope.new_child(f"scope", self.cs.rbp.value)
        return child



    def emit_scope_end(self, scope: CompilerScope) -> None:
        self.emitter.MOVRR(self.cs.rsp, self.cs.rbp)
        self.emitter.POP(self.cs.rbp)
        scope.discard()



    def emit_jump(self, opcode: int, label: int, scope: CompilerScope) -> None:
        if label is None:
            self.error(f"Label is None at {self.cs.rip.value}")
        scope.root.new_jump(self.cs.rip.value, label)
        if opcode == Opcode.JMP: self.emitter.JMP(label)
        elif opcode == Opcode.JMPZ: self.emitter.JMPZ(label)
        elif opcode == Opcode.JMPNZ: self.emitter.JMPNZ(label)
        elif opcode == Opcode.JMPGT: self.emitter.JMPGT(label)
        elif opcode == Opcode.JMPGTE: self.emitter.JMPGTE(label)
        elif opcode == Opcode.JMPLT: self.emitter.JMPLT(label)
        elif opcode == Opcode.JMPLTE: self.emitter.JMPLTE(label)
        elif opcode == Opcode.DJNZ: self.emitter.DJNZ(label)
        else: self.error(f"Unknown jump opcode {opcode} at {self.cs.rip.value}")

# expression helper functions.

    def var_to_stack(self, var: CTVariable) -> None:
        if var.scope.is_function_definition:
            self.emitter.LOADR(self.cs.rax, self.cs.rbp, var.stack_offset)
        else:
            self.emitter.LOAD(self.cs.rax, var.stack_offset)
        self.emitter.PUSH(self.cs.rax)                     # push rax

# compiling statements

    def compile_if_statement(self, statement: AstNodeIfStatement, scope: CompilerScope) -> None:
        self.emit_new_label("if", scope)
        self.compile_expression(statement.condition, scope)
        label_at_the_very_end_of_if = scope.get_new_label("label_end")
        label_for_JMPNZ_in_if = scope.get_new_label("label_jnz")
        self.emit_condition(label_for_JMPNZ_in_if, Opcode.JMPNZ, scope)
        self.compile_block(statement.block, scope)
        self.emit_jump(Opcode.JMP, label_at_the_very_end_of_if, scope)
        scope.set_label_address(label_for_JMPNZ_in_if, self.cs.rip.value)
        if statement.elif_block_list is not None:
            count = len(statement.elif_block_list)
            if count > 0:
                for elif_statement in statement.elif_block_list:
                    self.compile_elif_statement(elif_statement, label_at_the_very_end_of_if, scope)
        if statement.else_block is not None and len(statement.else_block.statements) > 0:
            self.compile_else_statement(statement.else_block, scope)
        scope.set_label_address(label_at_the_very_end_of_if, self.cs.rip.value)



    def compile_elif_statement(self, statement: AstNodeElifStatement, lbl_end: int, scope: CompilerScope) -> None:
        self.emit_new_label("elif", scope)
        self.compile_expression(statement.condition, scope)
        label = scope.root.get_new_label("elif_jnz")
        self.emit_condition(label, Opcode.JMPNZ, scope)
        self.compile_block(statement.statements, scope)
        self.emit_jump(Opcode.JMP, lbl_end, scope)
        scope.set_label_address(label, self.cs.rip.value)
        self.emit_new_label("/elif", scope)



    def compile_else_statement(self, statement: AstNodeElseStatement, scope: CompilerScope) -> None:
        self.emit_new_label("else", scope)
        self.compile_block(statement, scope)
        self.emit_new_label("/else", scope)



    def compile_foreach_statement(self, statement: AstNodeForeachStatement, scope: CompilerScope) -> None:
        foreach_end_label = scope.root.get_new_label("foreach_loop_end")
        self.emit_new_label("foreach_start", scope)
        foreach_continue_label = scope.get_new_label("foreach_loop_continue")
        foreach_scope = self.emit_scope(scope)
        # get the variable we are iterating over
        self.compile_expression(statement.expr, foreach_scope)  # array is on stack [rbp + 0]
        self.emitter.LEN(self.cs.rax)                              # len rax
        self.emitter.PUSH(self.cs.rax)                             # push length       [rbp + 1]
        self.emitter.XOR(self.cs.rax, self.cs.rax) # xor rax, rax
        self.emitter.PUSH(self.cs.rax)                             # push index        [rbp + 2]
        it = foreach_scope.define_variable("it", self.cs.rsp.value)
        it.node_type = AstNodetype.INTEGER
        it.value = 0
        self.emitter.PUSH(self.cs.rax)                             # push variable it  [rbp + 3]
        # preparation is done, we can start the loop:
        foreach_loop_start = self.emit_new_label("foreach_loop_start", scope)
        foreach_scope.continue_label = foreach_continue_label
        foreach_scope.break_label = foreach_end_label
        self.emitter.LOAD(self.cs.rax, 1 + foreach_scope.rbp_offset)                # load rax, [rbp + 1] # rax is now the length
        self.emitter.LOAD(self.cs.rbx, 2 + foreach_scope.rbp_offset)                # load rbx, [rbp + 2] # rbx is now index
        self.emitter.CMPRR(self.cs.rax, self.cs.rbx)       # cmprr rax, rbx
        self.emit_jump(Opcode.JMPZ, foreach_end_label, scope.root) # jmpz foreach_end_label if rax == rbx (we are at the end of the array)
        self.emitter.LOAD(self.cs.rax, 0 + foreach_scope.rbp_offset)                # load rax, [rbp + 0] # rax is now the array
        self.emitter.FEINDEX()              # rax = rax[rbx]      # rax is now the value for the variable it
        self.emitter.STOR(it.stack_offset, self.cs.rax)                # it is now the value of the array at the current index
        self.compile_block(statement.block, foreach_scope) # compile the block of the foreach statement.
        scope.set_label_address(foreach_continue_label, self.cs.rip.value)
        self.emitter.LOAD(self.cs.rax, 2 + foreach_scope.rbp_offset)                # load rax, [rbp + 2] # rax is now the index
        self.emitter.INC(self.cs.rax)                      # inc rax rax = rax + 1
        self.emitter.STOR(2 + foreach_scope.rbp_offset, self.cs.rax)                # stor [rbp + 2], rax increased the index
        self.emit_jump(Opcode.JMP, foreach_loop_start, scope)
        scope.set_label_address(foreach_end_label, self.cs.rip.value)
        self.emit_scope_end(foreach_scope)
        self.emit_new_label("foreach_end", scope)



    def compile_while_statement(self, statement: AstNodeWhileStatement, scope: CompilerScope) -> None:
        self.emit_new_label("while", scope)
        loop_start = scope.root.get_new_label("while_loop_start")
        loop_end   = scope.root.get_new_label("while_loop_end")
        while_scope = self.emit_scope(scope)
        while_scope.continue_label = loop_start
        while_scope.break_label = loop_end
        scope.set_label_address(loop_start, self.cs.rip.value)
        self.compile_expression(statement.condition, scope)
        self.emit_condition(loop_end, Opcode.JMPNZ, scope)
        self.compile_block(statement.block, while_scope)
        self.emit_jump(Opcode.JMP, loop_start, scope)
        scope.set_label_address(loop_end, self.cs.rip.value)
        self.emit_scope_end(while_scope)



    def compile_continue_statement(self, statement: AstNodeContinueStatement, scope: CompilerScope) -> None:
        if scope.continue_label is None:
            self.error(f"Continue statement not allowed at {statement.line}:{statement.column}")
        self.emit_new_label("continue", scope)
        self.emit_jump(Opcode.JMP, scope.continue_label, scope)



    def compile_break_statement(self, statement: AstNodeBreakStatement, scope: CompilerScope) -> None:
        if scope.break_label is None:
            self.error(f"Break statement not allowed at {statement.line}:{statement.column}")
        self.emit_new_label("break", scope)
        self.emit_jump(Opcode.JMP, scope.break_label, scope)



    def compile_return_statement(self, statement: AstNodeReturnStatement, scope: CompilerScope) -> None:
        if scope.exit_label is None:
            self.error(f"Return statement not allowed at {statement.line}:{statement.column}")
        self.emit_new_label("return", scope)
        if statement.expr is not None:
            self.compile_expression(statement.expr, scope)
            # the function return value is reserved at [rbp - 2]
            self.emitter.POP(self.cs.rax)                     # pop rax
            self.emitter.STORR(self.cs.rbp, -3, self.cs.rax)  # storr [rbp - 3], rax
        # jump to the exit label, the cleanup and the RET are executed there.
        self.emit_jump(Opcode.JMP, scope.exit_label, scope)



    def compile_select_statement(self, statement: AstNodeStatement, scope: CompilerScope) -> None:
        sc = MacalSelectCompiler(self)
        sc.compile_select(statement, scope)



    def compile_halt_statement(self, statement: AstNodeHaltStatement, scope: CompilerScope) -> None:
        self.emit_new_label("halt", scope)
        self.compile_expression(statement.expr, scope)
        self.emitter.POP(self.cs.rax)              # pop rax
        self.emitter.STOR(0, self.cs.rax)       # stor [0], rax -> rax is the exit code
        self.emitter.HALT()



    def compile_print_statement(self, statement: AstNodePrintStatement, scope: CompilerScope) -> None:
        self.emit_new_label("print", scope)
        for expr in statement.args:
            self.compile_expression(expr, scope)
            self.emitter.POP(self.cs.rax)              # pop rax
            self.emitter.PRNT()
        self.emitter.MOVRI(self.cs.rax, AstNodetype.STRING, "\n")
        self.emitter.PRNT()



    def compile_switch_statement(self, statement: AstNodeStatement, scope: CompilerScope) -> None:
        # A switch statement is basically a jump table.
        # So we need to compile the expression and then jump to the correct label.
        # If there is no match, we jump to the default label.
        switch_end_label = scope.root.get_new_label("switch_end")
        default_label = scope.root.get_new_label("switch_default")

        self.emit_new_label("switch", scope)
        self.compile_expression(statement.expr, scope)
        # The expression result is on the stack now this is the value that we need to match

        # Against each of the case values.
        # So first we build the jump table:
        jump_table = {}
        for case in statement.cases:
            case_label = scope.root.get_new_label("case")
            jump_table[case.case_value()] = case_label

        # Now we can compile the jump table.
        scope.switch_jump_tables.append(SwitchCaseTable(self.cs.rip.value, jump_table))
        self.emitter.MOVRI(self.cs.rax, AstNodetype.RECORD, jump_table)

        # Now we calculate which jump to make.
        self.emitter.POP(self.cs.rbx)              # pop rbx # rbx is the value to match
        self.emitter.HASFLDRR(self.cs.rax, self.cs.rbx) # hasfldrr rax, rbx. The ZF flag is set if the value is found.
        self.emit_jump(Opcode.JMPNZ, default_label, scope) # jmpz default_label if the value is not found.
        self.emitter.INDEX()
        self.emitter.JMPR(self.cs.rax) # jmp rax
        scope.set_label_address(default_label, self.cs.rip.value)
        if statement.default is not None:
            self.compile_block(statement.default.block, scope)
        self.emit_jump(Opcode.JMP, switch_end_label, scope)
        for case in statement.cases:
            scope.set_label_address(jump_table[case.case_value()], self.cs.rip.value)
            self.compile_block(case.block, scope)
            self.emit_jump(Opcode.JMP, switch_end_label, scope)
        scope.set_label_address(switch_end_label, self.cs.rip.value)



    def compile_is_type_statement(self, statement: AstNodeIsType, scope: CompilerScope) -> None:
        # This should cover all the IsXXXX type statements.
        self.emit_new_label("is_type", scope)
        self.compile_expression(statement.expr, scope)
        self.emitter.POP(self.cs.rax)              # pop rax
        self.emitter.CMPRTI(self.cs.rax, statement.TypeToCheck)
        self.emitter.SETZR(self.cs.rax)
        self.emitter.PUSH(self.cs.rax)              # push rax



    def compile_type_statement(self, statement: AstNodeTypeStatement, scope: CompilerScope) -> None:
        self.emit_new_label("Type", scope)
        self.compile_expression(statement.expr, scope)
        self.emitter.POP(self.cs.rax)              # pop rax
        self.emitter.TYPE(self.cs.rax)
        self.emitter.PUSH(self.cs.rax)              # push rax



    def compile_assignment_statement(self, statement: AstNodeAssignmentStatement, scope: CompilerScope) -> None:
        # Assignment works almost like a binary expression.
        # We are assured that the lhs is a variable.
        # The RHS can be anything.
        self.emit_new_label("assign", scope)
        var = scope.get_variable(statement.name)
        is_new_var = False
        if var is None: # we go assign a new var.
            #print("define variable: ", statement.name)
            #print("stack offset: ", self.cs.rsp.value)
            var = scope.define_variable(statement.name,self.cs.rsp.value)
            var.const = statement.const
            self.emitter.XOR(self.cs.rax, self.cs.rax) # xor rax, rax rax = 0
            self.emitter.PUSH(self.cs.rax)             # push rax reserve the variable on the stack.
            is_new_var = True
        if var.const and not is_new_var:
            self.error(f"Cannot assign to const {statement.name} at {statement.line}:{statement.column} scope({scope.name})")
        if not isinstance(var, CTVariable):
            self.error(f"Cannot assign to function {statement.name} at {statement.line}:{statement.column} scope({scope.name})")
        if statement.lhs.expr_type == AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            if is_new_var:
                self.error(f"Variable unassigned, can't read index at {statement.line}:{statement.column} scope({scope.name})")
            self.compile_indexed_variable_assignment(statement, var, scope)
            return
        func = self.compile_expression(statement.rhs, scope) # the value is on the stack now.
        if func is not None:
            self.scope.define_function_alias(statement.name, func.name)
        if  statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT:
            if statement.append is True:
                reg = self.cs.GetFreeRegister()
                self.emitter.POP(reg)                      # pop reg
                self.emitter.LOAD(self.cs.rax, var.stack_offset) # load rax, [rbp + stack_offset] rax is now the variable.
                self.emitter.APPEND(self.cs.rax, reg)        # append rax, reg
                self.emitter.STOR(var.stack_offset, self.cs.rax) # stor [rbp + stack_offset], rax
                self.cs.ReleaseRegister(reg)
                return
            self.emitter.POP(self.cs.rax)              # pop rax
            self.emitter.STOR(var.stack_offset, self.cs.rax) # stor [rbp + stack_offset], rax
            return
        # if it's not straight up assign, we need to do a binary operation with the variable value as the lhs, so we get it:
        self.emitter.LOAD(self.cs.rax, var.stack_offset) # load rax, [rbp + stack_offset] rax is now the variable.
        self.emitter.POP(self.cs.rbx)              # pop rbx
        if statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_INC:
            self.emitter.ADDRR(self.cs.rax, self.cs.rbx)        # addrr rax, rbx
        elif statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DEC:
            self.emitter.SUBRR(self.cs.rax, self.cs.rbx)        # subrr rax, rbx
        elif statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MUL:
            self.emitter.MULRR(self.cs.rax, self.cs.rbx)        # mulrr rax, rbx
        elif statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DIV:
            self.emitter.DIVRR(self.cs.rax, self.cs.rbx)        # divrr rax, rbx
        elif statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MOD:
            self.emitter.MODRR(self.cs.rax, self.cs.rbx)        # modrr rax, rbx
        elif statement.op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_POW:
            self.emitter.POWRR(self.cs.rax, self.cs.rbx)        # powrr rax, rbx
        self.emitter.STOR(var.stack_offset, self.cs.rax) # stor [rbp + stack_offset], rax



    # this is a continuation from compile_assign_statement where the variable is indexed.
    def compile_indexed_variable_assignment(self, statement: AstNodeAssignmentStatement, var: CTVariable, scope: CompilerScope) -> None:
        self.emit_new_label("assign_indexed", scope)
        if var.const:
            self.error(f"Cannot assign to const {statement.name} at {statement.line}:{statement.column} scope({scope.name})")
        self.var_to_stack(var)
        n = len(statement.lhs.index)
        if n > 1:
            for i in range(n-1):
                self.compile_expression(statement.lhs.index[i], scope)
                self.emitter.POP(self.cs.rbx)              # pop rbx
                self.emitter.POP(self.cs.rax)              # pop rax
                self.emitter.INDEX()        # rax = rax[rbx]
                self.emitter.PUSH(self.cs.rax)             # push rax
        # at this point there should be something on the top of the stack that is
        # working as a reference, so if we assign to that it'll change the original with it.
        self.compile_expression(statement.lhs.index[n-1], scope)
        # on stack is now:
        # [0] = value
        # [1] = index
        # now we need to handle the right hand side of the assignment.
        self.compile_assign_indexed(statement.op, var, statement.rhs, scope)


    # this is a continuation from compile_indexed_variable_assignment, it handles the actual assignment.
    def compile_assign_indexed(self, op: LexToken, var: CTVariable, rhs: AstNodeExpression, scope: CompilerScope) -> None:
        if var.const:
            self.error(f"Cannot assign to const {var.name} at {rhs.line}:{rhs.column} scope({scope.name})")
        self.emit_new_label("assign_indexed_cnt", scope)
        func = self.compile_expression(rhs, scope) # the value is on the stack now.
        if func is not None:
            self.scope.define_function_alias(var.name, func.name)
        reg: BytecodeRegister = self.cs.GetFreeRegister()
        if  op.token_type == AstNodetype.OPERATOR_ASSIGNMENT:
            self.emitter.POP(reg)                      # pop reg
            self.emitter.POP(self.cs.rbx)              # pop rbx [index]
            self.emitter.POP(self.cs.rax)              # pop rax [value]
            self.emitter.INDEXR(reg)       # indexr rax[rbx] = reg
            self.cs.ReleaseRegister(reg)
            return
        self.emitter.POP(reg)                      # pop reg => value to assign
        self.emitter.POP(self.cs.rbx)              # pop rbx [index]
        self.emitter.POP(self.cs.rax)              # pop rax [value]
        # rax[rbx] = LHS
        # reg      = RHS
        # We need to put rax and rbx back on the stack, because we need them later on.
        self.emitter.PUSH(self.cs.rax)              # push rax
        self.emitter.PUSH(self.cs.rbx)              # push rbx
        # Now get the value, so we can do something with it:
        self.emitter.INDEX()  # pop rax = rax[rbx]
        self.emit_movrr(self.cs.rbx, reg) # movrr rbx = reg
        if op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_INC:
            self.emitter.ADDRR(self.cs.rax, self.cs.rbx)        # addrr rax, rbx
        elif op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DEC:
            self.emitter.SUBRR(self.cs.rax, self.cs.rbx)        # subrr rax, rbx
        elif op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MUL:
            self.emitter.MULRR(self.cs.rax, self.cs.rbx)        # mulrr rax, rbx
        elif op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_DIV:
            self.emitter.DIVRR(self.cs.rax, self.cs.rbx)        # divrr rax, rbx
        elif op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_MOD:
            self.emitter.MODRR(self.cs.rax, self.cs.rbx)        # modrr rax, rbx
        elif op.token_type == AstNodetype.OPERATOR_ASSIGNMENT_POW:
            self.emitter.POWRR(self.cs.rax, self.cs.rbx)        # powrr rax, rbx
        self.emit_movrr(reg, self.cs.rax) # movrr reg = rax
        # reg is the resulting value, that we need to write as if it was a normal assign.
        self.emitter.POP(self.cs.rbx)              # pop rbx [index]
        self.emitter.POP(self.cs.rax)              # pop rax [value]
        self.emitter.INDEXR(reg)       # indexr rax[rbx] = reg
        self.cs.ReleaseRegister(reg)
        self.emit_new_label("assign_indexed_cnt_end", scope)



    def has_params(self, params: List[AstNodeFunctionParameter]) -> bool:
        return any(param.node_type == AstNodetype.PARAMS for param in params)



    def compile_function_arguments(self, statement: AstNodeFunctionCallStatement, func: CTFunction, scope: CompilerScope) -> None:
        self.emit_new_label("function_arguments", scope)
        if len(statement.args) != len(func.parameters) and not self.has_params(func.parameters):
            self.error(f"Function {statement.name} at {statement.line}:{statement.column} scope({scope.name}) requires {len(func.parameters)} arguments, {len(statement.args)} given")
        for expr in statement.args:
            self.emit_new_label(f"function_argument: {expr}", scope)
            self.compile_expression(expr, scope)
        self.emitter.MOVRI(self.cs.rax, AstNodetype.INTEGER, len(statement.args)) # function arguments count is in rax
        self.emitter.PUSH(self.cs.rax)              # push rax



    def compile_clean_function_arguments(self, statement: AstNodeFunctionCallStatement) -> None:
        # gets the return value from the stack, puts it in a temp reg, then pops all the arguments
        # from the stack and puts the return value back on the stack.
        # but only if the function had any arguments.
        reg = self.cs.GetFreeRegister()
        self.emitter.POP(reg) # get the return value from the stack.
        if len(statement.args) > 0:
            for _ in range(len(statement.args)): # pop all the arguments from the stack.
                self.emitter.POP(self.cs.rax)
        self.emitter.POP(self.cs.rax) # pop the number of arguments value from the stack.
        self.emitter.PUSH(reg) #push the return value onto the stack again.
        self.cs.ReleaseRegister(reg)



    def compile_function_call(self, statement: AstNodeFunctionCallStatement, scope: CompilerScope) -> None:
        self.emit_new_label("function_call", scope)
        func = scope.get_function(statement.name)
        if func is None:
            var = scope.get_variable(statement.name)
            if var is not None:
                self.compile_variable_function_call(statement, var, scope)
                return
            self.error(f"Function {statement.name} not found at {statement.line}:{statement.column} scope({scope.name})")
        self.compile_function_arguments(statement, func, scope)
        self.emitter.CALL(func.start_address)
        self.compile_clean_function_arguments(statement)
        self.emitter.POP(self.cs.rax) # pop rax -> RAX has function return value.


    # this is a continuation of compile_function_call where the identifier is a variable instead of a function.
    def compile_variable_function_call(self, statement: AstNodeFunctionCallStatement, var: CTVariable, scope: CompilerScope) -> None:
        self.emit_new_label("variable_function_call", scope)
        # TODO: Make it so that it also works for indexed variable access.
        # It now only works based on the variable name, not on the index.
        # This bit is almost the same as it's the signature of the function call.
        #self.cs.write_string(var.name)
        func = scope.get_function_by_alias(var.name)
        if func is None:
            self.error(f"Function {statement.name} not found at {statement.line}:{statement.column} scope({scope.name})")
        self.compile_function_arguments(statement, func, scope)
        # So we are a variable, we need to figure out if we are a function pointer.
        self.emitter.LOAD(self.cs.rax, var.stack_offset) # load rax, [rbp + stack_offset] rax is now the variable.
        # rax is now the value of the variable.
        if statement.index is not None: # ok, we sneaked in an indexed one under the radar :P
            var.index = statement.index
            self.emitter.PUSH(self.cs.rax)              # push rax
            for i in range(len(var.index)):
                self.compile_expression(var.index[i], scope) # this pushes the result of the expression.
                self.emitter.POP(self.cs.rbx)              # pop rbx
                self.emitter.POP(self.cs.rax)              # pop rax
                self.emitter.INDEX()        # rax = rax[rbx]
                self.emitter.PUSH(self.cs.rax)             # push rax
            self.emitter.POP(self.cs.rax)              # pop rax
            self.cs.rax.value_type = AstNodetype.FUNCTION # this is needed, because inside the array this information was lost.
            self.emitter.SETRTI(self.cs.rax, AstNodetype.FUNCTION)
        # validate if it is a function pointer.
        self.emitter.CMPRTI(self.cs.rax, AstNodetype.FUNCTION) # compare register.type with immediate type
        self.emitter.ERRORNZ(f"Variable {var.name} is not a function pointer. Scope({scope.name})") # error out when zero flag is not set.
        # if we get here, we have a valid function pointer in RAX
        # so we can call it.
        # warning cannot use rax in callr, callr destroys rax.
        self.emitter.MOVRR(self.cs.rbx, self.cs.rax)
        self.emitter.CALLR(self.cs.rbx)                # callr rbx
        self.compile_clean_function_arguments(statement)
        # The downside of this is that we can't change the location in memory for a function.
        # Unless we do some very elaborate work with setting the value of the variable to the correct address before
        # this is ran. But that is a problem for later.



    def compile_function_definition(self, statement: AstNodeFunctionDefinition, scope: CompilerScope) -> None:
        func = scope.define_function(statement.name, self.cs.rip.value+1) # +9 for the jump instruction.
        if func is None:
            self.error(f"Function {statement.name} already defined at {statement.line}:{statement.column} scope({scope.name})")
        function_label = scope.get_new_label(func.name)
        function_call_exit_label = scope.root.get_new_label(f"{func.name}_exit")
        function_call_end_label = scope.root.get_new_label(f"{func.name}_end")
        self.emit_jump(Opcode.JMP, function_call_end_label, scope) # functions go in between the other code, so we must jump over them.
        scope.set_label_address(function_label, self.cs.rip.value)
        # make function scope.
        function_scope = self.emit_scope(scope)
        function_scope.is_function_definition = True
        function_scope.can_search_parent = False
        param_count = len(statement.parameters)
        for i, param in enumerate(statement.parameters):
            func.parameters.append(param)
            offset = FUNCTION_CALL_PARAMETER_STACK_OFFSET - (param_count-1) + i
            v = function_scope.define_variable(param.name, offset)
            v.node_type = param.node_type

        function_scope.exit_label = function_call_exit_label
        if statement.return_type is not None:
            func.return_type = statement.return_type
            function_scope.define_variable("return_value", FUNCTION_CALL_RETURN_VALUE_STACK_OFFSET)
        if statement.external:
            self.compile_external_function_definition(statement, func)
        else:
            self.compile_block(statement.body, function_scope)
        self.compile_function_definition_end(function_call_exit_label, function_call_end_label, function_scope)



    def compile_external_function_definition(self, statement: AstNodeFunctionDefinition, func: CTFunction) -> None:
        func.external = True
        func.module = statement.module
        func.function = statement.function
        self.emitter.CALLE(func.module, func.function, len(statement.parameters))
        self.emitter.STORR(self.cs.rbp, -3, self.cs.rax)  # storr [rbp - 3], rax



    def compile_function_definition_end(self, function_call_exit_label: int, function_call_end_label: int, scope: CompilerScope) -> None:
        scope.set_label_address(function_call_exit_label, self.cs.rip.value)
        # exit label here.
        self.emit_scope_end(scope)
        self.emitter.RET()          # ret
        # because rbp is popped after resetting the stack pointer,
        # the return value of the function is now on top of the stack.
        scope.set_label_address(function_call_end_label, self.cs.rip.value)

    # Meta data compilation

    def compile_datasegment(self, scope: CompilerScope) -> None:
        metadata = {
            "VARIABLES": [var.name for var in scope.variables],
            "FUNCTIONS": [func.name for func in scope.functions],
            "LIBRARIES": [lib for lib in scope.libraries],
            "LABELS": [(label.name, label.address) for label in scope.labels],
            "JUMPS": [(jump.address, jump.label) for jump in scope.jump_table],
            "RESERVED": []
        }
        self.emitter.DATASEG(metadata)

    # expression compilation

    def compile_expression(self, expression: AstNodeExpression, scope: CompilerScope) -> Optional[CTFunction]:
        self.emit_new_label("expression", scope) # up for commenting out, it's too chatty.
        if expression.expr_type == AstNodetype.LITERAL_EXPRESSION:
            self.compile_literal_expression(expression, scope)
        elif expression.expr_type == AstNodetype.UNARY_EXPRESSION:
            self.compile_unary_expression(expression, scope)
        elif expression.expr_type == AstNodetype.BINARY_EXPRESSION:
            self.compile_binary_expression(expression, scope)
        elif expression.expr_type == AstNodetype.VARIABLE_EXPRESSION:
            return self.compile_variable_expression(expression, scope)
        elif expression.expr_type == AstNodetype.INDEXED_VARIABLE_EXPRESSION:
            self.compile_indexed_variable_expression(expression, scope)
        elif expression.expr_type == AstNodetype.FUNCTION_CALL_EXPRESSION:
            return self.compile_function_call_expression(expression, scope)
        elif expression.expr_type == AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION:
            self.compile_variable_function_call_expression(expression, scope)
        elif expression.expr_type == AstNodetype.IS_TYPE_STATEMENT:
            self.compile_is_type_statement(expression, scope)
        elif expression.expr_type == AstNodetype.TYPE_STATEMENT:
            self.compile_type_statement(expression, scope)
        else:
            self.error(f"Unknown expression type {expression.expr_type} at {expression.line}:{expression.column}")


    def compile_literal_expression(self, expression: AstNodeLiteralExpression, scope: CompilerScope) -> None:
        # A literal expression means a literal value gets written to the stack.
        self.emit_new_label("literal_expression", scope)
        if expression.value_type == AstNodetype.NEW_ARRAY:
            expression.value_type = AstNodetype.ARRAY
            expression.value = []
        self.emitter.MOVRI(self.cs.rax, expression.value_type, expression.value)
        self.emitter.PUSH(self.cs.rax)



    def compile_unary_expression(self, expression: AstNodeUnaryExpression, scope: CompilerScope) -> None:
        self.emit_new_label("unary_expression", scope)
        self.compile_expression(expression.right, scope)
        self.emitter.POP(self.cs.rax)              # pop rax
        if expression.op.token_type == AstNodetype.OPERATOR_SUBTRACTION:
            self.emitter.NEG(self.cs.rax)
        elif expression.op.token_type == AstNodetype.NOT_STATEMENT:
            self.emitter.NOT(self.cs.rax)
        elif expression.op.token_type == AstNodetype.BITWISE_NOT:
            self.error(f"Bitwise not not implemented at {expression.line}:{expression.column}")
        self.emitter.PUSH(self.cs.rax)



    def compile_binary_expression(self, expression: AstNodeBinaryExpression, scope: CompilerScope) -> None:
        self.emit_new_label("binary_expression", scope)
        # binary expression is simple. First we compile the left and right side of the expression
        # and then we pop the values in RAX and RBX apply the operator and then push RAX back onto the stack.
        self.compile_expression(expression.left, scope)
        self.compile_expression(expression.right, scope)
        self.emitter.POP(self.cs.rbx)              # pop rbx rhs
        self.emitter.POP(self.cs.rax)              # pop rax lhs
        if expression.op.token_type == AstNodetype.OPERATOR_ADDITION:
            self.emitter.ADDRR(self.cs.rax, self.cs.rbx)   # addrr rax, rbx
        elif expression.op.token_type == AstNodetype.OPERATOR_SUBTRACTION:
            self.emitter.SUBRR(self.cs.rax, self.cs.rbx)   # subrr rax, rbx
        elif expression.op.token_type == AstNodetype.OPERATOR_MULTIPLICATION:
            self.emitter.MULRR(self.cs.rax, self.cs.rbx)   # mulrr rax, rbx
        elif expression.op.token_type == AstNodetype.OPERATOR_DIVISION:
            self.emitter.DIVRR(self.cs.rax, self.cs.rbx)   # divrr rax, rbx
        elif expression.op.token_type == AstNodetype.OPERATOR_POWER:
            self.emitter.POWRR(self.cs.rax, self.cs.rbx)   # powrr rax, rbx
        elif expression.op.token_type == AstNodetype.OPERATOR_MODULUS:
            self.emitter.MODRR(self.cs.rax, self.cs.rbx)   # modrr rax, rbx
        elif expression.op.token_type  == AstNodetype.COMPARETOR_EQUAL:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETZR(self.cs.rax)
        elif expression.op.token_type  == AstNodetype.COMPARETOR_NOT_EQUAL:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETNZR(self.cs.rax)
        elif expression.op.token_type  == AstNodetype.COMPARETOR_LESS_THAN:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETLR(self.cs.rax)
        elif expression.op.token_type  == AstNodetype.COMPARETOR_LESS_THAN_EQUAL:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETLER(self.cs.rax)
        elif expression.op.token_type  == AstNodetype.COMPARETOR_GREATER_THAN:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETGR(self.cs.rax)
        elif expression.op.token_type  == AstNodetype.COMPARETOR_GREATER_THAN_EQUAL:
            self.emitter.CMPRR(self.cs.rax, self.cs.rbx)
            self.emitter.SETGER(self.cs.rax)
        elif expression.op.token_type == AstNodetype.AND_STATEMENT:
            self.emitter.ANDRR(self.cs.rax, self.cs.rbx)
        elif expression.op.token_type == AstNodetype.OR_STATEMENT:
            self.emitter.ORRR(self.cs.rax, self.cs.rbx)
        elif expression.op.token_type == AstNodetype.XOR_STATEMENT:
            self.emitter.XOR(self.cs.rax, self.cs.rbx)
        self.emitter.PUSH(self.cs.rax)              # push rax



    def compile_variable_expression(self, expression: AstNodeVariableExpression, scope: CompilerScope) -> Optional[CTFunction]:
        self.emit_new_label("variable_expression", scope)
        var = scope.get_variable(expression.name)
        if var is None:
            func = scope.get_function(expression.name)
            if func is not None:
                self.compile_function_expression(func)
                return func
            self.error(f"Variable or function {expression.name} not found at {expression.line}:{expression.column} scope({scope.name})")
        if not isinstance(var, CTVariable):
            self.error (f"Variable {expression.name} is not a variable at {expression.line}:{expression.column} scope({scope.name})")
        # variables live on the stack, the start of the stack for a scope is in the rbp register.
        # so we read the value from [rbp + var.offset] (our stack starts at 0 and counts up)
        # and push it to the stack.
        self.var_to_stack(var)



    def compile_indexed_variable_expression(self, expression: AstNodeIndexedVariableExpression, scope: CompilerScope) -> None:
        self.emit_new_label("indexed_variable_expression", scope)
        # Indexed variable expression is very hard.
        # Getting the variable is not a big deal, but handling the index is.
        var = scope.get_variable(expression.name)
        if var is None:
            self.error(f"Variable {expression.name} not defined at {expression.line}:{expression.column} scope({scope.name})")
        if not isinstance(var, CTVariable):
            self.error(f"Variable {expression.name} is not a variable at {expression.line}:{expression.column} scope({scope.name})")
        self.var_to_stack(var)
        # The variable value on the stack.
        # Now we need to get the index value.
        for expr in expression.index:
            self.compile_expression(expr, scope)    # the index value is on the stack now.
            self.emitter.POP(self.cs.rbx)              # pop rbx
            self.emitter.POP(self.cs.rax)              # pop rax
            self.emitter.INDEX()        # rax = rax[rbx]
            self.emitter.PUSH(self.cs.rax)             # push rax



    def compile_function_expression(self, func: CTFunction) -> None:
        # This should only happen in an assignment.
        self.emit_new_label("function_expression", self.scope)
        self.emitter.MOVRI(self.cs.rax, AstNodetype.FUNCTION, func.start_address)
        self.emitter.PUSH(self.cs.rax)



    def compile_function_call_expression(self, expression: AstNodeFunctionCallExpression, scope: CompilerScope) -> None:
        self.emit_new_label("function_call_expression", scope)
        func = scope.get_function(expression.name)
        if func is None: # maybe it's a variable? don't know, it should be handled by the other expr types.
            self.error(f"Function {expression.name} not found at {expression.line}:{expression.column} scope({scope.name})")
        self.compile_function_arguments(expression, func, scope)
        self.emitter.CALL(func.start_address)
        self.compile_clean_function_arguments(expression)



    def compile_variable_function_call_expression(self, expression: AstNodeVariableFunctionCallExpression, scope: CompilerScope) -> None:
        self.emit_new_label("variable_function_call_expression", scope)
        var = scope.get_variable(expression.variable.name)
        if var is None:
            self.error(f"Variable {expression.variable.name} not defined at {expression.line}:{expression.column} scope({scope.name})")
        for expr in expression.args:
            self.compile_expression(expr)
        func = scope.get_function_by_alias(expression.variable.name)
        self.emitter.CALL(func.start_address)
        self.compile_clean_function_arguments(expression)

    
# Include implementation

    def find_library(self, name: str) -> Optional[str]:
        for path in SearchPath:
            lib_path = os.path.join(path, f"{name}.mcl")
            if os.path.exists(lib_path):
                return lib_path
        return None



    def load_library(self, path: str) -> Optional[str]:
        if not os.path.exists(path):
            return None
        with open(path, "r") as f:
            return f.read()



    def compile_include_statement(self, statement: AstNodeIncludeStatement, scope: CompilerScope) -> None:
        self.emit_new_label("include_statement", scope)
        for lib in statement.libraries:
            if lib.name in scope.libraries:
                continue
            lib_path = self.find_library(lib.name)
            if lib_path is None:
                self.error(f"Library {lib.name} not found at {lib.line}:{lib.column}")
            source = self.load_library(lib_path)
            if source is None:
                self.error(f"Library {lib.name} not found at {lib.line}:{lib.column}")
            self.emit_new_label(f"include_{lib.name}", scope)
            self.scope.libraries[lib.name] = self.cs.rip.value
            lex = Lexer(source)
            parser = Parser(lex.tokenize(), lib_path)
            self.compile_block(parser.parse(), scope)

# end of file :)



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the AST node for the SWITCH statment.

from __future__ import annotations
from typing import Optional, List
from .ast_nodetype import AstNodetype
from .lex_token import LexToken
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_node_case_statement import AstNodeCaseStatement
from .ast_node_default_statement import AstNodeDefaultStatement


class AstNodeSwitchStatement(AstNodeStatement):
    def __init__(self, token: LexToken, expr: AstNodeExpression) -> AstNodeSwitchStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.SWITCH_STATEMENT
        self.expr: AstNodeExpression = expr
        self.cases: List[AstNodeCaseStatement] = []
        self.default: Optional[AstNodeDefaultStatement] = None


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# This is the ast node for the while statement.

from __future__ import annotations
from typing import Optional
from .ast_node_expression import AstNodeExpression
from .ast_node_statement import AstNodeStatement
from .ast_node_block import AstNodeBlock
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeWhileStatement(AstNodeStatement):
    def __init__(self, token: LexToken) -> AstNodeWhileStatement:
        super().__init__(token)
        self.expr_type = AstNodetype.WHILE_STATEMENT
        self.condition: Optional[AstNodeExpression] = None
        self.block: Optional[AstNodeBlock] = None

    def __str__(self) -> str:
        return f"WhileStatement({self.condition}, {self.block})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}WhileStatement:")
        if self.condition is not None:
            self.condition.tree(indent + '    ')
        if self.block is not None:
            self.block.tree(indent + '    ')



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the statement node for the AST.

from __future__ import annotations
from .ast_node import AstNode
from .lex_token import LexToken

class AstNodeStatement(AstNode):
    def __init__(self, token: LexToken) -> AstNodeStatement:
        super().__init__()
        self.token: LexToken = token
        self.expr_type = token.token_type

    @property
    def name(self) -> str:
        return self.token.value

    @property
    def line(self) -> int:
        return self.token.lineno

    @property
    def column(self) -> int:
        return self.token.offset

    def __str__(self) -> str:
        return f"AstNodeStatement({self.expr_type}, {self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}Statement({self.expr_type}, {self.name})")



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the function call expression node for the AST.

from __future__ import annotations
from typing import List
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeFunctionCallExpression(AstNodeExpression):
    def __init__(self, token: LexToken, args: List[AstNodeExpression]) -> AstNodeFunctionCallExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.FUNCTION_CALL_EXPRESSION
        self.args: List[AstNodeExpression] = args

    @property
    def name(self) -> str:
        return self._token.value

    def __str__(self) -> str:
        return f"AstNodeFunctionCallExpression({self.name})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}{self}")
        if self.args:
            for arg in self.args:
                arg.tree(indent + '    ')
        else:
            print()



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# Decompiler for Macal VM instructionset.

from __future__ import annotations
from .macal_instructions import MacalInstructionList, MacalInstruction
from .conversion import convertToHexAddr
import pyximport; pyximport.install()
from .cmacal_vm import MacalVm
from .ast_nodetype import AstNodetype

class MacalDecompiler:
    def __init__(self, filename: str, verbose: bool = False) -> MacalDecompiler:
        self.vm: MacalVm = MacalVm(filename)
        self.vm.rip.value = 0
        self.instructions: MacalInstructionList = MacalInstructionList()
        self.metadata: dict = {}
        self.labels: list = []
        self.verbose: bool = verbose
        if filename is not None:
            lm = len(self.vm.memory) - 1
            data = []
            if self.vm.memory[lm][0] == 77:
                data = self.vm.memory[lm]
            elif self.vm.memory[lm-1][0] == 77:
                data = self.vm.memory[lm-1]
            if data != []:
                self.metadata = data[1]
                self.labels = self.metadata['LABELS']



    def __PrintLabels(self, addr: int) -> None:
        flag = False
        for label in self.labels:
            if label[1] == addr:
                print(f"{label[0]}:", end=' ')
                flag = True
        if flag is True:
            print()



    def Decompile(self):
        addr = 0
        print()
        print(f"Decompiling: {self.vm.filename}")
        print()
        print("Address       Instruction")
        while addr < len(self.vm.memory):
            self.__PrintLabels(addr)
            opcode = self.vm.memory[addr]
            op: MacalInstruction = self.instructions.fromOpcode(opcode[0])
            print(f"{convertToHexAddr(addr)}    {op.Name:<6} ", end=' ')
            n = 1
            for operand in op.Operands:
                if n > 1:
                    print(", ", end='')
                if operand == "addr":
                    print(convertToHexAddr(opcode[n]), end='')
                elif operand == "[addr]":
                    print(f"[{convertToHexAddr(opcode[n])}]", end='')
                elif operand.startswith("reg"):
                    print(f"{self.vm.opcode_to_register_map[opcode[n]].name}", end='')
                elif operand == "imm":
                    if opcode[0] != 80:
                        value = f"{opcode[n+1]}"
                        if opcode[n] == AstNodetype.STRING:
                            value = value.replace('\n', '\\n')	# escape newlines
                            value = value.replace('\r', '\\r')	# escape cr
                            value = f'"{value}"'
                        print(f"{opcode[n].name} {value.strip()}", end='')
                    else:
                        print(f"{opcode[n]}", end='')
                    n += 1
                elif operand == "errmsg" or operand == "module" or operand == "function" or operand == "paramcount" or operand == "offset":
                    print(opcode[n], end='')
                elif operand == "metadata":
                    print("METADATA", end='')
                elif operand == "type":
                    print(opcode[n].name, end='')
                elif operand == "str":
                    print(f'"{opcode[n]}"', end='')
                else:
                    raise Exception(f"Unknown operand: {operand}")
                n+= 1
            print()
            addr += 1
        print()
        print(f"Finished decompiling: {self.vm.filename}")
        print()




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This dictionary table maps the statement type to the corresponding statement lexeme.

from .ast_nodetype import AstNodetype

StatementTypeTable = {
    "const": AstNodetype.CONST_STATEMENT,
    "if": AstNodetype.IF_STATEMENT,
    "else": AstNodetype.ELSE_STATEMENT,
    "elif": AstNodetype.ELIF_STATEMENT,
    "foreach": AstNodetype.FOREACH_STATEMENT,
    "while": AstNodetype.WHILE_STATEMENT,
    "continue": AstNodetype.CONTINUE_STATEMENT,
    "break": AstNodetype.BREAK_STATEMENT,
    "return": AstNodetype.RETURN_STATEMENT,
    "select": AstNodetype.SELECT_STATEMENT,
    "distinct": AstNodetype.DISTINCT_STATEMENT,
    "as": AstNodetype.AS_STATEMENT,
    "from": AstNodetype.FROM_STATEMENT,
    "where": AstNodetype.WHERE_STATEMENT,
    "order": AstNodetype.ORDER_STATEMENT,
    "merge": AstNodetype.MERGE_STATEMENT,
    "into": AstNodetype.INTO_STATEMENT,
    "halt": AstNodetype.HALT_STATEMENT,
    "include": AstNodetype.INCLUDE_STATEMENT,
    "external": AstNodetype.EXTERNAL_STATEMENT,
    "print": AstNodetype.PRINT_STATEMENT,
    "switch": AstNodetype.SWITCH_STATEMENT,
    "case": AstNodetype.CASE_STATEMENT,
    "default": AstNodetype.DEFAULT_STATEMENT,
    "false": AstNodetype.BOOLEAN,
    "true": AstNodetype.BOOLEAN,
    "nil": AstNodetype.NIL,
    "and": AstNodetype.AND_STATEMENT,
    "or": AstNodetype.OR_STATEMENT,
    "not": AstNodetype.NOT_STATEMENT,
    "xor": AstNodetype.XOR_STATEMENT,
    "integer": AstNodetype.INTEGER,
    "float": AstNodetype.FLOAT,
    "string": AstNodetype.STRING,
    "variable": AstNodetype.VARIABLE,
    "function": AstNodetype.FUNCTION,
    "bool": AstNodetype.BOOLEAN,
    "array": AstNodetype.ARRAY,
    "record": AstNodetype.RECORD,
    "params": AstNodetype.PARAMS,
    "Type": AstNodetype.TYPE_STATEMENT,
    "IsString": AstNodetype.IS_STRING_STATEMENT,
    "IsInt": AstNodetype.IS_INT_STATEMENT,
    "IsFloat": AstNodetype.IS_FLOAT_STATEMENT,
    "IsBool": AstNodetype.IS_BOOL_STATEMENT,
    "IsRecord": AstNodetype.IS_RECORD_STATEMENT,
    "IsArray": AstNodetype.IS_ARRAY_STATEMENT,
    "IsFunction": AstNodetype.IS_FUNCTION_STATEMENT,
    "IsNil": AstNodetype.IS_NIL_STATEMENT,

}


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the ast node for assigning a value to a variable.

from __future__ import annotations
from typing import Union
from .ast_node import AstNode
from .ast_node_variable_expression import AstNodeVariableExpression
from .ast_node_expression import AstNodeExpression
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_node_statement import AstNodeStatement
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

# This is now almost identical to a binary expression, except that the lhs is known to be a variable expression.
class AstNodeAssignmentStatement(AstNodeStatement):
    def __init__(self, var_expr: AstNodeVariableExpression, op: LexToken, expr: AstNodeExpression) -> AstNodeAssignmentStatement:
        super().__init__(var_expr._token)
        self.expr_type = AstNodetype.ASSIGNMENT_STATEMENT
        self.lhs = var_expr
        self.variable_name: str = var_expr.name
        self.op: LexToken = op
        self.rhs: Union[AstNodeLiteralExpression, AstNodeExpression]
        if isinstance(expr, list):
            self.rhs = AstNodeLiteralExpression(LexToken(AstNodetype.ARRAY, expr, 0 , 0, 0), expr, AstNodetype.ARRAY)
        elif isinstance(expr, dict):
            self.rhs = AstNodeLiteralExpression(LexToken(AstNodetype.RECORD, expr, 0 , 0, 0), expr, AstNodetype.RECORD)
        else:
            self.rhs: AstNodeExpression = expr
        self.append: bool = False
        self.const: bool = False

    def __str__(self) -> str:
        return f"AstNodeAssignmentStatement({self.variable_name}, {self.op}, {self.lhs} {self.rhs})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}AssignmentStatement():")
        self.lhs.tree(indent + '    ')
        print(f"{indent}    {self.op}")
        if isinstance(self.rhs, AstNode):
            self.rhs.tree(indent + '    ')
        else:
          print(indent, '    ', self.rhs)




#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the Macal decompiler, it will decompile a .mbc file into bytecode mnenomics.

import os
import argparse
from typing import Optional
from macal.__about__ import __version__, __author__, __author_email__, __license__
from macal.config import SearchPath
from macal.macal_decompiler import MacalDecompiler

bytecode_extension  = '.mbc'

def SetupSearchPath() -> None:
    search_path = [os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext")]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)

def check_file(file) -> Optional[str]:
    if file is None: return
    if not file.endswith(bytecode_extension):
        if os.path.exists(f"{file}{bytecode_extension}"):
            file = f"{file}{bytecode_extension}"
        else:
            raise Exception("Invalid Macal bytecode file.")
    return file

def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description='Macal decompiler')
    parser.add_argument('file', nargs='?', type=check_file, help='Macal file')
    parser.add_argument('-s', '--verbose', action='store_true', help='Verbose mode')
    parser.add_argument('-v', '--version', action='version', version=f'Macal 2 interpreter v{__version__}')
    return parser

def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.file is None:
        parser.print_help()
        return
    if args.verbose:
        PrintBanner()
    MacalDecompiler(args.file).Decompile()


def PrintBanner() -> None:
    print(f"Macal decompiler v{__version__}")
    print(f"Author:  {__author__}")
    print(f"Email:   {__author_email__}")
    print(f"License: {__license__}")
    print()

def PrintTokens(tokens: list) -> None:
    print("Tokens:")
    for token in tokens:
        print(token)
    print()

def PrintAst(ast) -> None:
    print("AST:")
    ast.tree()
    print()

if __name__ == "__main__":
    Run()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      09-11-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the new Macal Runtime, it interpret/execute a compiled Macal (.mbc) file.

import os
import argparse
from .__about__ import __version__, __author__, __author_email__, __license__
from .config import SearchPath
import pyximport; pyximport.install()
from .cmacal_vm import MacalVm

bytecode_extension  = '.mbc'

def SetupSearchPath():
    search_path = [os.path.dirname(__file__),
        os.path.join(os.path.dirname(__file__), "lib"),
        os.path.join(os.path.dirname(__file__), "lib", "ext"),
        os.getcwd(),
        os.path.join(os.getcwd(), "lib"),
        os.path.join(os.getcwd(), "lib", "ext")]
    for path in search_path:
        if path not in SearchPath:
            SearchPath.append(path)

def check_file(file) -> str:
    if file is None: return
    if not file.endswith(bytecode_extension):
        if os.path.exists(f"{file}{bytecode_extension}"):
            file = f"{file}{bytecode_extension}"
        else:
            raise Exception("Invalid Macal bytecode file.")
    return file

def BuildArgParser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description='Macal runtime')
    parser.add_argument('file', nargs='?', type=check_file, help='Macal bytecode file (.mbc)')
    parser.add_argument('-v', '--version', action='version', version=f'Macal 2 runtime v{__version__}')
    return parser


def ShowResult(exitCode: int) -> None:
    if exitCode == 0:
        print("Process exits OK")
    else:
        print(f"Process exits with error: {exitCode}")
    exit(exitCode)

def Run():
    SetupSearchPath()
    parser = BuildArgParser()
    args = parser.parse_args()
    if args.file is None:
        parser.print_help()
        return
    vm = MacalVm(args.file)
    vm.Execute()
    ShowResult(vm.stack[0][1])

if __name__ == "__main__":
    Run()


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the AST node for a function call expression through a variable.

from __future__ import annotations
from typing import List
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .ast_node_variable_expression import AstNodeVariableExpression

class AstNodeVariableFunctionCallExpression(AstNodeExpression):
    def __init__(self, variable: AstNodeVariableExpression, args: List[AstNodeExpression]) -> AstNodeVariableFunctionCallExpression:
        super().__init__(variable._token)
        self.expr_type = AstNodetype.VARIABLE_FUNCTION_CALL_EXPRESSION
        self.variable: AstNodeVariableExpression = variable
        self.args: List[AstNodeExpression] = args

    def __str__(self) -> str:
        return f"VariableFunctionCallExpression({self.variable}, {self.args})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}VariableFunctionCallExpression:")
        self.variable.tree(indent + '    ')
        for arg in self.args:
            arg.tree(indent + '    ')


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the indexed variable access node for the AST.

from __future__ import annotations
from typing import List
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeIndexedVariableExpression(AstNodeExpression):
    def __init__(self, token: LexToken) -> AstNodeIndexedVariableExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.INDEXED_VARIABLE_EXPRESSION
        self.name: LexToken = token.value
        self.index: List[AstNodeExpression] = []

    def __str__(self) -> str:
        return f"AstNodeIndexedVariableExpression({self.name}, {self.index})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}AstNodeIndexedVariableExpression:")
        print(f"{indent}    {self.name}{self.index}")



#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#


# AST Node base class

from __future__ import annotations
from .ast_nodetype import AstNodetype

class AstNode:
    def __init__(self) -> AstNode:
        self.expr_type: AstNodetype = AstNodetype.UNDEFINED

    def __str__(self) -> str:
        return f"AstNode()"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}AstNode:")
        print(f"{indent}    {self.expr_type}")


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the binary expression node for the AST.

from __future__ import annotations
from .ast_node_expression import AstNodeExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeBinaryExpression(AstNodeExpression):
    def __init__(self, op: LexToken, left: AstNodeExpression, right: AstNodeExpression) -> AstNodeBinaryExpression:
        super().__init__(op)
        self.expr_type = AstNodetype.BINARY_EXPRESSION
        self.left: AstNodeExpression = left
        self.right: AstNodeExpression = right
        self.op: LexToken = op

    def __str__(self) -> str:
        return f"AstNodeExpression({self.left}, {self.op}, {self.right})"

    def __repr__(self) -> str:
        return self.__str__()

    def tree(self, indent: str = '') -> None:
        print(f"{indent}BinaryExpression:")
        print(f"{indent}    {self._token.value} ({self._token.token_type}) {self.line}, {self.column}")
        self.left.tree(indent + '    ')
        self.right.tree(indent + '    ')


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This is the node object for an array literal expression.

from __future__ import annotations
from .ast_node_literal_expression import AstNodeLiteralExpression
from .ast_nodetype import AstNodetype
from .lex_token import LexToken

class AstNodeLiteralArrayExpression(AstNodeLiteralExpression):
    def __init__(self, token: LexToken, array: list) -> AstNodeLiteralArrayExpression:
        super().__init__(token)
        self.expr_type = AstNodetype.LITERAL_ARRAY_EXPRESSION
        self.array: list = array

    def tree(self, indent: str = '') -> None:
        print(f"{indent}LiteralArrayExpression:")
        print(f'    {indent}', end='')
        for item in self.array:
            print(item, end=', ')
        print()

    def value(self) -> list:
        return self.array

    def __str__(self) -> str:
        return f"AstNodeLiteralArrayExpression({self.token})"

    def __repr__(self) -> str:
        return self.__str__()


# Product:   Macal 2
# Author:    Marco Caspers
# Date:      16-10-2023
#

"""Macal time library implementation"""

from datetime import datetime
import time


NUM_SECONDS_FIVE_MINUTES = 300
NUM_SECONDS_ONE_HOUR = 3600
TIME_FORMAT = '%Y%m%d%H%M%S'
ISO_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
ISO_TIME_tzFORMAT = "%Y-%m-%dT%H:%M:%S.0Z"
__stopwatch__ = 0

def DateToUnix(var: str) -> int:
    """Convert a date_string of format YYYYMMDDhhmmss to unix time integer.
       Assumes the date string object is UTC time."""
    dt = datetime.strptime(var, TIME_FORMAT)
    epoch = datetime(1970, 1, 1)
    return int((dt - epoch).total_seconds())

def IsoToUnix(var: str) -> int:
    """Convert a date_string of format %Y-%m-%dT%H:%M:%S.%f to unix time integer.
       Assumes the date string object is in iso format."""
    dt = datetime.strptime(var, ISO_TIME_FORMAT)
    epoch = datetime(1970, 1, 1)
    return int((dt - epoch).total_seconds())

def DateFromUnix(var: int) -> str:
    """Converts time in seconds since UNIX EPOCH to UTC Time format"""
    return time.strftime(TIME_FORMAT, time.gmtime(var))

def IsoFromUnix(var: str) -> None:
    """Converts time in seconds since UNIX EPOCH to UTC Time format"""
    return time.strftime(ISO_TIME_tzFORMAT, time.gmtime(var))

def UtcNow() -> str:
    return datetime.utcnow().strftime("%Y%m%d%H%M%S")

def UtcIsoNow() -> str:
    return "{}Z".format(datetime.utcnow().isoformat())

def IsoNow() -> str:
    return datetime.now().isoformat()

def Now() -> str:
    return datetime.now().strftime("%Y%m%d%H%M%S")

def PerfCounter() -> float:
    return time.perf_counter()




# This is the supporting library for Macal 2.0 keyrings.

import keyring

def GetPassword(username):
    return keyring.get_password("passwords", username)

def SetPassword(username, password):
    keyring.set_password("passwords", username, password)

def DeletePassword(username):
    keyring.delete_password("passwords", username)

def SetMerakiApiKey(customer, key):
    keyring.set_password("meraki", customer, key)

def GetMerakiApiKey(customer):
    return keyring.get_password("meraki", customer)


#
# Product:   Macal
# Author:    Marco Caspers
# Email:     SamaDevTeam@westcon.com
# License:   MIT License
# Date:      24-10-2023
#
# Copyright 2023 Westcon-Comstor
#

# This module contains the conversion utilities for the Macal language.

from typing import Any, Optional

from .ast_nodetype import AstNodetype
from .lex_token import LexToken

def convertToHex(value: int, digits: int = 4, prefix: str = '') -> str:
    if isinstance(value, int):
        return f'{prefix}{value:0{digits}X}'
    return f'{value}'.strip()

def typeFromOpcode(opcode: int) -> Optional[AstNodetype]:
    return next((type for type in AstNodetype if type.value == opcode), None)

def isType(opcode: int) -> bool:
    for type in AstNodetype:
        if type.value == opcode:
            return True
    return False

def convertToValue(token: LexToken) -> Any:
    if token.token_type == AstNodetype.INTEGER:
        return int(token.value)
    elif token.token_type == AstNodetype.FLOAT:
        return float(token.value)
    elif (token.token_type == AstNodetype.STRING
            or token.token_type == AstNodetype.STRING_INTERPOLATION_STRING_PART
            or token.token_type == AstNodetype.STRING_INTERPOLATION_END):
        #return token.value
        return token.value[1:-1]
    elif token.token_type == AstNodetype.BOOLEAN:
        return token.value == 'true'
    elif token.token_type == AstNodetype.BOOLEAN_TRUE:
        return True
    elif token.token_type == AstNodetype.BOOLEAN_FALSE:
        return False
    elif token.token_type == AstNodetype.NIL:
        return None
    else:
        raise Exception(f"convertToValue(): Unknown type {token.token_type}")

def compareTypes(type1: AstNodetype, type2: AstNodetype) -> bool:
    if type1 == type2:
        return True
    # handle the oddball cases
    if type1 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type1 == AstNodetype.STRING_INTERPOLATION_END:
        type1 = AstNodetype.STRING
    if type2 == AstNodetype.STRING_INTERPOLATION_STRING_PART or type2 == AstNodetype.STRING_INTERPOLATION_END:
        type2 = AstNodetype.STRING
    return type1 == type2

def operandToStr(type: AstNodetype) -> str:
    if type == AstNodetype.OPERATOR_ADDITION:
        return '+'
    elif type == AstNodetype.OPERATOR_SUBTRACTION:
        return '-'
    elif type == AstNodetype.OPERATOR_MULTIPLICATION:
        return '*'
    elif type == AstNodetype.OPERATOR_DIVISION:
        return '/'
    elif type == AstNodetype.OPERATOR_MODULUS:
        return '%'
    elif type == AstNodetype.OPERATOR_POWER:
        return '^'
    elif type == AstNodetype.COMPARETOR_EQUAL:
        return '=='
    elif type == AstNodetype.COMPARETOR_NOT_EQUAL:
        return '!='
    elif type == AstNodetype.COMPARETOR_LESS_THAN:
        return '<'
    elif type == AstNodetype.COMPARETOR_LESS_THAN_EQUAL:
        return '<='
    elif type == AstNodetype.COMPARETOR_GREATER_THAN:
        return '>'
    elif type == AstNodetype.COMPARETOR_GREATER_THAN_EQUAL:
        return '>='
    elif type == AstNodetype.AND_STATEMENT:
        return 'and'
    elif type == AstNodetype.OR_STATEMENT:
        return 'or'
    elif type == AstNodetype.NOT_STATEMENT:
        return 'not'
    elif type == AstNodetype.XOR_STATEMENT:
        return 'xor'
    elif type == AstNodetype.BITWISE_AND:
        return '&'
    elif type == AstNodetype.BITWISE_OR:
        return '|'
    elif type == AstNodetype.BITWISE_NOT:
        return '~'
    elif type == AstNodetype.OPERATOR_ASSIGNMENT:
        return '='
    elif type == AstNodetype.OPERATOR_INCREMENT:
        return '++'
    elif type == AstNodetype.OPERATOR_DECREMENT:
        return '--'
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_INC:
        return '+='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_DEC:
        return '-='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_MUL:
        return '*='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_DIV:
        return '/='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_MOD:
        return '%='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_POW:
        return '^='
    elif type == AstNodetype.OPERATOR_ASSIGNMENT_NEG:
        return '~='
    else:
        return 'unknown'

def typeFromValue(value: Any) -> AstNodetype:
    # BOOLEAN BEFORE INTEGER as BOOLEAN is a subclass of INTEGER
    if isinstance(value, bool):
        return AstNodetype.BOOLEAN
    elif isinstance(value, int):
        return AstNodetype.INTEGER
    elif isinstance(value, float):
        return AstNodetype.FLOAT
    elif isinstance(value, str):
        return AstNodetype.STRING
    elif value is None or value == 'nil' or value == AstNodetype.NIL:
        return AstNodetype.NIL
    elif isinstance(value, list):
        return AstNodetype.ARRAY
    elif isinstance(value, dict):
        return AstNodetype.RECORD
    else:
        raise Exception(f"typeFromValue(): Unknown type {type(value)}")

def convertToHexAddr(addr: int) -> str:
    return f'0x{addr:08X}'

def typeToStr(type: AstNodetype):
    return type.name



# This is a test for creating an agent style appliaction that uses the Macal interpreter/runtime.

import os
import argparse
import sys
from macal.macal_vm import MacalVm

AGENTVERSIONMSG = "Meraki Agent v10.0.rc1"

gl_orgname = "orgname"
gl_orgid = "orgid"
gl_apikey = "api"
gl_hostname = "host"
gl_script = "script"
gl_configuration = {
    "bandwidth_timespan": "86400",
    "enable_bandwidth": "True",
    "enable_demo": "False",
    "enable_logging": "True",
    "enable_public_ip": "True",
    "show_address": "True",
    "show_errors": "True",
    "show_firmware": "True",
    "show_soft_errors": "True",
    "show_execution_time": "True",
    "save_unknown_events": "True",
    "security_event_timespan": "86400",
    "syslog_address": "",
    "device_serial": "",
    "customers" : [
        {"name": "Stage Entertainment", "orgid": "549236", "apikey": "a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6"},
        {"name": "Medialane", "orgid": "549236", "apikey": "a1b2c3d4e5f6g7h8i9j0k1l2m3n4o5p6"},	
    ]
}

def ParseArgs():
    """Parse and return the command line arguments."""
    parser = argparse.ArgumentParser(description='Run Meraki Agent script .mcl file.')
    
    parser.add_argument('-c',  '--configure',      help='Change the configuration',                           action='store_true')
    parser.add_argument('-d',  '--del_customer',   help='Deletes a customer from the configuration',          action='store_true')
    parser.add_argument('-e',  '--execute',        help='Load and execute the *.pyql configuration file.',    action='store_true')
    parser.add_argument('-io', '--ignore_org',     help='Ignore the org parameter value, can be used for special scripts only.',    action='store_true')
    parser.add_argument('-l',  '--list_customers', help='Shows a list of all customers in the configuration', action='store_true')
    parser.add_argument('-m',  '--check_customer', help='Checks if a customer exists in the configuration',   action='store_true')
    parser.add_argument('-n',  '--new_customer',   help='Creates a new customer in the configuration',        action='store_true')
    parser.add_argument('-so', '--safetyoverride', help='Overrides exeption handling.',         					 action="store_true")
    parser.add_argument('-sc', '--show_config',    help='Shows the current agent configuration',              action="store_true")
    parser.add_argument('-v',  '--version',        help="Shows agent version information.",                   action='version', version=AGENTVERSIONMSG)
    parser.add_argument('-vv', '--versionverbose', help='Shows verbose agent version information.',           action="store_true")
    
    reqd = parser.add_argument_group('required named arguments')

    op_delcust    = '-d'  in sys.argv or '--del_customer'   in sys.argv
    op_exec       = '-e'  in sys.argv or '--execute'        in sys.argv
    op_newcust    = '-n'  in sys.argv or '--new_customer'   in sys.argv
    op_checkcust  = '-m'  in sys.argv or '--check_customer' in sys.argv
    op_vversion   = '-vv' in sys.argv or '--versionverbose' in sys.argv
    op_orgname    = '-o'  in sys.argv or '--orgname'        in sys.argv

    req_hostname  = op_exec
    req_apikey    = op_exec or op_vversion
    req_orgname   = op_delcust or op_exec or op_newcust or op_checkcust
    req_orgid     = (op_delcust or op_exec or op_newcust or op_checkcust) and not op_orgname

    reqd.add_argument('-H', '--hostname', help='Hostname variable. Used by API', required = req_hostname)
    reqd.add_argument('-k', '--apikey',   help='API Key to access the API', required = req_apikey)
    reqd.add_argument('-o', '--orgname',  help='Orginasation/Customer name variable. Used by API and configuration.', required = req_orgname)
    reqd.add_argument('-i', '--orgid',    help='Orginasation/Customer ID variable. Used by API and configuration.', required = req_orgid)
    reqd.add_argument('-s', '--script',   help='The PyQL script to execute.', required = op_exec)

    #For configuration -c // --config, all optional.
    reqd.add_argument('-bc', '--backup_config',           help='Make a backup of the current configuration to this file.', required = False)
    reqd.add_argument('-bt', '--bandwidth_timespan',      help='Set network bandwidth timespan for a customer.', required = False)
    reqd.add_argument('-eb', '--enable_bandwidth',        help='Enable/disable showing bandwidth history and trafficshaping max bandwidth for a customer.', required = False)
    reqd.add_argument('-ed', '--enable_demo',             help='Enable/disable demo mode for a customer.', required = False)
    reqd.add_argument('-el', '--enable_logging',          help='Enable/disable logging.', required = False)
    reqd.add_argument('-ep', '--enable_public_ip',        help='Enable/disable showing public ip for a customer.', required = False)
    reqd.add_argument('-rb', '--restore_config',          help='Restore configuration from this file.', required = False)
    reqd.add_argument('-sa', '--show_address',            help='Enable/disable show address for a customer.', required = False)
    reqd.add_argument('-se', '--show_errors',             help='Enable/disable show errors.', required = False)
    reqd.add_argument('-sf', '--show_firmware',           help='Enable/disable show firmware for a customer.', required = False)
    reqd.add_argument('-sse', '--show_soft_errors',       help='Enable/disable show soft errors (script execution).', required = False)
    reqd.add_argument('-st', '--show_execution_time',     help='Enable/disable show execution time.', required = False)
    reqd.add_argument('-su', '--save_unknown_events',     help='Enable/disable Save unknown network events.', required = False)
    reqd.add_argument('-sv', '--security_event_timespan', help='Set security event timespan for a customer.', required = False)
    reqd.add_argument('-sy', '--syslog_address',          help='Hostname syslog server.', required = False)
    reqd.add_argument('-sn', '--device_serial',           help='Device serial number for scripts.', required = False)
    return parser.parse_args()


def main():
    args = ParseArgs()
    global gl_hostname, gl_apikey, gl_orgname, gl_orgid, gl_script
    vm = MacalVm("usereserved.mbc")
    vm.SetReservedVariable("orgname", args.orgname)
    vm.SetReservedVariable("orgid", args.orgid)
    vm.SetReservedVariable("apikey", args.apikey)
    vm.SetReservedVariable("hostname", args.hostname)
    vm.SetReservedVariable("script", args.script)
    vm.SetReservedVariable("configuration", gl_configuration)
    result = vm.Execute()
    if result != -2:
        print(f"Runtime Error: expected -2 from execute, got: {result}.")
    exitcode = vm.stack[0][1]
    print("Process exitcode: ", exitcode)
    return exitcode


if __name__ == "__main__":
    main()